#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "track_to_new_artifact",
    "track_to_existing_artifact",
    "get_artifact_attribute",
    "list_artifact_files",
]

from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Type,
)

from bravado.exception import HTTPNotFound

from neptune.common.backends.utils import with_api_exceptions_handler
from neptune.exceptions import (
    ArtifactNotFoundException,
    ArtifactUploadingError,
    FetchAttributeNotFoundException,
    NeptuneEmptyLocationException,
)
from neptune.internal.artifacts.file_hasher import FileHasher
from neptune.internal.artifacts.types import (
    ArtifactDriver,
    ArtifactDriversMap,
    ArtifactFileData,
)
from neptune.internal.backends.api_model import (
    ArtifactAttribute,
    ArtifactModel,
)
from neptune.internal.backends.swagger_client_wrapper import SwaggerClientWrapper
from neptune.internal.operation import (
    AssignArtifact,
    Operation,
)
from neptune.internal.utils.paths import path_to_str


def _compute_artifact_size(artifact_file_list: List[ArtifactFileData]):
    artifact_size = 0
    for artifact_file in artifact_file_list:
        if artifact_file.size is None:
            # whole artifact's size is undefined in this case
            return None
        artifact_size += artifact_file.size
    return artifact_size


def _filter_empty_directory_files(files: List[ArtifactFileData]) -> List[ArtifactFileData]:
    return list(filter(lambda x: not _is_s3_empty_directory_file(x), files))


def _is_s3_empty_directory_file(file: ArtifactFileData) -> bool:
    return file.type == "S3" and file.size == 0


def track_to_new_artifact(
    swagger_client: SwaggerClientWrapper,
    project_id: str,
    path: List[str],
    parent_identifier: str,
    entries: List[Tuple[str, Optional[str]]],
    default_request_params: Dict,
    exclude_directory_files: bool,
    exclude_metadata_from_hash: bool,
) -> Optional[Operation]:
    files: List[ArtifactFileData] = _extract_file_list(path, entries)

    if exclude_directory_files:
        files = _filter_empty_directory_files(files)

    if not files:
        raise ArtifactUploadingError("Uploading an empty Artifact")

    artifact_hash = (
        _compute_artifact_hash_without_metadata(files) if exclude_metadata_from_hash else _compute_artifact_hash(files)
    )

    artifact = create_new_artifact(
        swagger_client=swagger_client,
        project_id=project_id,
        artifact_hash=artifact_hash,
        parent_identifier=parent_identifier,
        size=_compute_artifact_size(files),
        default_request_params=default_request_params,
    )

    if not artifact.received_metadata:
        upload_artifact_files_metadata(
            swagger_client=swagger_client,
            project_id=project_id,
            artifact_hash=artifact_hash,
            files=files,
            default_request_params=default_request_params,
        )

    return AssignArtifact(path=path, hash=artifact_hash)


def track_to_existing_artifact(
    swagger_client: SwaggerClientWrapper,
    project_id: str,
    path: List[str],
    artifact_hash: str,
    parent_identifier: str,
    entries: List[Tuple[str, Optional[str]]],
    default_request_params: Dict,
    exclude_directory_files: bool,
) -> Optional[Operation]:
    files: List[ArtifactFileData] = _extract_file_list(path, entries)

    if exclude_directory_files:
        files = _filter_empty_directory_files(files)

    if not files:
        raise ArtifactUploadingError("Uploading an empty Artifact")

    artifact = create_artifact_version(
        swagger_client=swagger_client,
        project_id=project_id,
        artifact_hash=artifact_hash,
        parent_identifier=parent_identifier,
        files=files,
        default_request_params=default_request_params,
    )

    return AssignArtifact(path=path, hash=artifact.hash)


def _compute_artifact_hash_without_metadata(files: List[ArtifactFileData]) -> str:
    return FileHasher.get_artifact_hash_without_metadata(files)


def _compute_artifact_hash(files: List[ArtifactFileData]) -> str:
    return FileHasher.get_artifact_hash(files)


def _extract_file_list(path: List[str], entries: List[Tuple[str, Optional[str]]]) -> List[ArtifactFileData]:
    files: List[ArtifactFileData] = list()

    for entry_path, entry_destination in entries:
        driver: Type[ArtifactDriver] = ArtifactDriversMap.match_path(entry_path)
        artifact_files = driver.get_tracked_files(path=entry_path, destination=entry_destination)

        if len(artifact_files) == 0:
            raise NeptuneEmptyLocationException(location=entry_path, namespace="/".join(path))

        files.extend(artifact_files)

    return files


@with_api_exceptions_handler
def create_new_artifact(
    swagger_client: SwaggerClientWrapper,
    project_id: str,
    artifact_hash: str,
    parent_identifier: str,
    size: int,
    default_request_params: Dict,
) -> ArtifactModel:
    params = {
        "projectIdentifier": project_id,
        "hash": artifact_hash,
        "size": size,
        "parentIdentifier": parent_identifier,
        **add_artifact_version_to_request_params(default_request_params),
    }
    try:
        result = swagger_client.api.createNewArtifact(**params).response().result
        return ArtifactModel(
            hash=result.artifactHash,
            received_metadata=result.receivedMetadata,
            size=result.size,
        )
    except HTTPNotFound:
        raise ArtifactNotFoundException(artifact_hash)


@with_api_exceptions_handler
def upload_artifact_files_metadata(
    swagger_client: SwaggerClientWrapper,
    project_id: str,
    artifact_hash: str,
    files: List[ArtifactFileData],
    default_request_params: Dict,
) -> ArtifactModel:
    params = {
        "projectIdentifier": project_id,
        "hash": artifact_hash,
        "artifactFilesDTO": {"files": [ArtifactFileData.to_dto(a) for a in files]},
        **add_artifact_version_to_request_params(default_request_params),
    }
    try:
        result = swagger_client.api.uploadArtifactFilesMetadata(**params).response().result
        return ArtifactModel(
            hash=result.artifactHash,
            size=result.size,
            received_metadata=result.receivedMetadata,
        )
    except HTTPNotFound:
        raise ArtifactNotFoundException(artifact_hash)


@with_api_exceptions_handler
def create_artifact_version(
    swagger_client: SwaggerClientWrapper,
    project_id: str,
    artifact_hash: str,
    parent_identifier: str,
    files: List[ArtifactFileData],
    default_request_params: Dict,
) -> ArtifactModel:
    params = {
        "projectIdentifier": project_id,
        "hash": artifact_hash,
        "parentIdentifier": parent_identifier,
        "artifactFilesDTO": {"files": [ArtifactFileData.to_dto(a) for a in files]},
        **add_artifact_version_to_request_params(default_request_params),
    }
    try:
        result = swagger_client.api.createArtifactVersion(**params).response().result
        return ArtifactModel(
            hash=result.artifactHash,
            size=result.size,
            received_metadata=result.receivedMetadata,
        )
    except HTTPNotFound:
        raise ArtifactNotFoundException(artifact_hash)


@with_api_exceptions_handler
def get_artifact_attribute(
    swagger_client: SwaggerClientWrapper,
    parent_identifier: str,
    path: List[str],
    default_request_params: Dict,
) -> ArtifactAttribute:
    requests_params = add_artifact_version_to_request_params(default_request_params)
    params = {
        "experimentId": parent_identifier,
        "attribute": path_to_str(path),
        **requests_params,
    }
    try:
        result = swagger_client.api.getArtifactAttribute(**params).response().result
        return ArtifactAttribute(hash=result.hash)
    except HTTPNotFound:
        raise FetchAttributeNotFoundException(path_to_str(path))


@with_api_exceptions_handler
def list_artifact_files(
    swagger_client: SwaggerClientWrapper,
    project_id: str,
    artifact_hash: str,
    default_request_params: Dict,
) -> List[ArtifactFileData]:
    requests_params = add_artifact_version_to_request_params(default_request_params)
    params = {
        "projectIdentifier": project_id,
        "hash": artifact_hash,
        **requests_params,
    }
    try:
        result = swagger_client.api.listArtifactFiles(**params).response().result
        return [ArtifactFileData.from_dto(a) for a in result.files]
    except HTTPNotFound:
        raise ArtifactNotFoundException(artifact_hash)


def add_artifact_version_to_request_params(default_request_params: Dict) -> Dict:
    current_artifact_version = "2"

    return {
        "_request_options": {
            **default_request_params["_request_options"],
            "headers": {
                **default_request_params["_request_options"]["headers"],
                "X-Neptune-Artifact-Api-Version": current_artifact_version,
            },
        }
    }



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["SyncRunner"]

import logging
import os
import threading
import time
from pathlib import Path
from typing import (
    Iterable,
    List,
    Optional,
    Sequence,
)

from neptune.cli.abstract_backend_runner import AbstractBackendRunner
from neptune.cli.utils import (
    get_metadata_container,
    get_offline_dirs,
    get_project,
    get_qualified_name,
    is_container_synced_and_remove_junk,
    iterate_containers,
    split_dir_name,
)
from neptune.common.exceptions import NeptuneConnectionLostException
from neptune.constants import (
    ASYNC_DIRECTORY,
    OFFLINE_DIRECTORY,
    OFFLINE_NAME_PREFIX,
)
from neptune.envs import NEPTUNE_SYNC_BATCH_TIMEOUT_ENV
from neptune.exceptions import CannotSynchronizeOfflineRunsWithoutProject
from neptune.internal.backends.api_model import (
    ApiExperiment,
    Project,
)
from neptune.internal.container_type import ContainerType
from neptune.internal.disk_queue import DiskQueue
from neptune.internal.id_formats import (
    QualifiedName,
    UniqueId,
)
from neptune.internal.operation import Operation
from neptune.internal.operation_processors.operation_storage import OperationStorage
from neptune.internal.utils.logger import logger

retries_timeout = int(os.getenv(NEPTUNE_SYNC_BATCH_TIMEOUT_ENV, "3600"))


class SyncRunner(AbstractBackendRunner):
    def sync_container(self, container_path: Path, experiment: ApiExperiment) -> None:
        qualified_container_name = get_qualified_name(experiment)
        logger.info("Synchronising %s", qualified_container_name)
        for execution_path in container_path.iterdir():
            self.sync_execution(
                execution_path=execution_path,
                container_id=experiment.id,
                container_type=experiment.type,
            )
        logger.info("Synchronization of %s %s completed.", experiment.type.value, qualified_container_name)

    def sync_execution(
        self,
        execution_path: Path,
        container_id: UniqueId,
        container_type: ContainerType,
    ) -> None:
        operation_storage = OperationStorage(execution_path)
        with DiskQueue(
            dir_path=operation_storage.data_path,
            to_dict=lambda x: x.to_dict(),
            from_dict=Operation.from_dict,
            lock=threading.RLock(),
        ) as disk_queue:
            while True:
                batch = disk_queue.get_batch(1000)
                if not batch:
                    break
                version = batch[-1].ver
                batch = [element.obj for element in batch]

                start_time = time.monotonic()
                expected_count = len(batch)
                version_to_ack = version - expected_count
                while True:
                    try:
                        processed_count, _ = self._backend.execute_operations(
                            container_id=container_id,
                            container_type=container_type,
                            operations=batch,
                            operation_storage=operation_storage,
                        )
                        version_to_ack += processed_count
                        batch = batch[processed_count:]
                        disk_queue.ack(version)
                        if version_to_ack == version:
                            break
                    except NeptuneConnectionLostException as ex:
                        if time.monotonic() - start_time > retries_timeout:
                            raise ex
                        logger.warning(
                            "Experiencing connection interruptions."
                            " Will try to reestablish communication with Neptune."
                            " Internal exception was: %s",
                            ex.cause.__class__.__name__,
                        )

    def sync_all_registered_containers(self, base_path: Path) -> None:
        async_path = base_path / ASYNC_DIRECTORY
        for container_type, unique_id, path in iterate_containers(async_path):
            if not is_container_synced_and_remove_junk(path):
                container = get_metadata_container(
                    backend=self._backend,
                    container_id=unique_id,
                    container_type=container_type,
                )
                if container:
                    self.sync_container(container_path=path, experiment=container)

    def sync_selected_registered_containers(
        self, base_path: Path, qualified_container_names: Sequence[QualifiedName]
    ) -> None:
        for name in qualified_container_names:
            container = get_metadata_container(
                backend=self._backend,
                container_id=name,
            )
            if container:
                container_path = base_path / ASYNC_DIRECTORY / f"{container.type.create_dir_name(container.id)}"
                container_path_deprecated = base_path / ASYNC_DIRECTORY / f"{container.id}"
                if container_path.exists():
                    self.sync_container(container_path=container_path, experiment=container)
                elif container_path_deprecated.exists():
                    self.sync_container(container_path=container_path_deprecated, experiment=container)
                else:
                    logger.warning("Warning: Run '%s' does not exist in location %s", name, base_path)

    def _register_offline_container(self, project: Project, container_type: ContainerType) -> Optional[ApiExperiment]:
        try:
            if container_type == ContainerType.RUN:
                return self._backend.create_run(project.id)
            else:
                raise ValueError("Only runs are supported in offline mode")
        except Exception as e:
            logger.warning(
                "Exception occurred while trying to create a run" " on the Neptune server. Please try again later",
            )
            logging.exception(e)
            return None

    @staticmethod
    def _move_offline_container(
        base_path: Path,
        offline_dir: str,
        server_id: UniqueId,
        server_type: ContainerType,
    ) -> None:
        online_dir = server_type.create_dir_name(container_id=server_id)
        # create async directory for run
        (base_path / ASYNC_DIRECTORY / online_dir).mkdir(parents=True)
        # mv offline directory inside async one
        (base_path / OFFLINE_DIRECTORY / offline_dir).rename(
            base_path / ASYNC_DIRECTORY / online_dir / "exec-0-offline"
        )

    def register_offline_containers(
        self, base_path: Path, project: Project, offline_dirs: Iterable[str]
    ) -> List[ApiExperiment]:
        result = []
        for offline_dir in offline_dirs:
            offline_path = base_path / OFFLINE_DIRECTORY / offline_dir
            if offline_path.is_dir():
                container_type, _ = split_dir_name(dir_name=offline_dir)
                container = self._register_offline_container(project, container_type=container_type)
                if container:
                    self._move_offline_container(
                        base_path=base_path,
                        offline_dir=offline_dir,
                        server_id=container.id,
                        server_type=container.type,
                    )
                    logger.info("Offline container %s registered as %s", offline_dir, get_qualified_name(container))
                    result.append(container)
            else:
                logger.warning("Offline container %s not found on disk.", offline_dir)
        return result

    def sync_offline_containers(
        self,
        base_path: Path,
        project_name: Optional[QualifiedName],
        offline_dirs: Sequence[UniqueId],
    ):
        if offline_dirs:
            project = get_project(project_name, backend=self._backend)
            if not project:
                raise CannotSynchronizeOfflineRunsWithoutProject
            registered_containers = self.register_offline_containers(base_path, project, offline_dirs)
            offline_containers_names = [get_qualified_name(exp) for exp in registered_containers]
            self.sync_selected_registered_containers(base_path, offline_containers_names)

    def sync_all_offline_containers(self, base_path: Path, project_name: QualifiedName) -> None:

        offline_dirs = get_offline_dirs(base_path)
        self.sync_offline_containers(base_path, project_name, offline_dirs)

    def sync_selected_containers(
        self,
        base_path: Path,
        project_name: Optional[str],
        container_names: Sequence[str],
    ) -> None:
        non_offline_container_names = [
            QualifiedName(name) for name in container_names if not name.startswith(OFFLINE_NAME_PREFIX)
        ]
        self.sync_selected_registered_containers(base_path, non_offline_container_names)

        offline_dirs = [
            UniqueId(name[len(OFFLINE_NAME_PREFIX) :])
            for name in container_names
            if name.startswith(OFFLINE_NAME_PREFIX)
        ]
        self.sync_offline_containers(base_path, project_name, offline_dirs)

    def sync_all_containers(self, base_path: Path, project_name: Optional[str]) -> None:
        self.sync_all_registered_containers(base_path)
        self.sync_all_offline_containers(base_path, project_name)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "ARTIFACT_ATTRIBUTE_SPACE",
    "DIFF_HEAD_INDEX_PATH",
    "LOG_ATTRIBUTE_SPACE",
    "MONITORING_ATTRIBUTE_SPACE",
    "MONITORING_STDERR_ATTRIBUTE_PATH",
    "MONITORING_STDOUT_ATTRIBUTE_PATH",
    "MONITORING_TRACEBACK_ATTRIBUTE_PATH",
    "PARAMETERS_ATTRIBUTE_SPACE",
    "PROPERTIES_ATTRIBUTE_SPACE",
    "SOURCE_CODE_ATTRIBUTE_SPACE",
    "SOURCE_CODE_ENTRYPOINT_ATTRIBUTE_PATH",
    "SOURCE_CODE_FILES_ATTRIBUTE_PATH",
    "SYSTEM_ATTRIBUTE_SPACE",
    "SYSTEM_DESCRIPTION_ATTRIBUTE_PATH",
    "SYSTEM_HOSTNAME_ATTRIBUTE_PATH",
    "SYSTEM_NAME_ATTRIBUTE_PATH",
    "SYSTEM_STATE_ATTRIBUTE_PATH",
    "SYSTEM_TAGS_ATTRIBUTE_PATH",
    "SYSTEM_FAILED_ATTRIBUTE_PATH",
    "SYSTEM_STAGE_ATTRIBUTE_PATH",
    "SIGNAL_TYPE_STOP",
    "SIGNAL_TYPE_ABORT",
    "UPSTREAM_INDEX_DIFF",
]

ARTIFACT_ATTRIBUTE_SPACE = "artifacts/"

LOG_ATTRIBUTE_SPACE = "logs/"

MONITORING_ATTRIBUTE_SPACE = "monitoring/"
MONITORING_STDERR_ATTRIBUTE_PATH = f"{MONITORING_ATTRIBUTE_SPACE}stderr"
MONITORING_STDOUT_ATTRIBUTE_PATH = f"{MONITORING_ATTRIBUTE_SPACE}stdout"
MONITORING_TRACEBACK_ATTRIBUTE_PATH = f"{MONITORING_ATTRIBUTE_SPACE}traceback"

PARAMETERS_ATTRIBUTE_SPACE = "parameters/"

PROPERTIES_ATTRIBUTE_SPACE = "properties/"

SOURCE_CODE_ATTRIBUTE_SPACE = "source_code/"
SOURCE_CODE_ENTRYPOINT_ATTRIBUTE_PATH = f"{SOURCE_CODE_ATTRIBUTE_SPACE}entrypoint"
SOURCE_CODE_FILES_ATTRIBUTE_PATH = f"{SOURCE_CODE_ATTRIBUTE_SPACE}files"

SYSTEM_ATTRIBUTE_SPACE = "sys/"
SYSTEM_DESCRIPTION_ATTRIBUTE_PATH = f"{SYSTEM_ATTRIBUTE_SPACE}description"
SYSTEM_HOSTNAME_ATTRIBUTE_PATH = f"{SYSTEM_ATTRIBUTE_SPACE}hostname"
SYSTEM_NAME_ATTRIBUTE_PATH = f"{SYSTEM_ATTRIBUTE_SPACE}name"
SYSTEM_STATE_ATTRIBUTE_PATH = f"{SYSTEM_ATTRIBUTE_SPACE}state"
SYSTEM_TAGS_ATTRIBUTE_PATH = f"{SYSTEM_ATTRIBUTE_SPACE}tags"
SYSTEM_FAILED_ATTRIBUTE_PATH = f"{SYSTEM_ATTRIBUTE_SPACE}failed"
SYSTEM_STAGE_ATTRIBUTE_PATH = f"{SYSTEM_ATTRIBUTE_SPACE}stage"

SIGNAL_TYPE_STOP = "neptune/stop"
SIGNAL_TYPE_ABORT = "neptune/abort"

DIFF_HEAD_INDEX_PATH = f"{SOURCE_CODE_ATTRIBUTE_SPACE}diff"
UPSTREAM_INDEX_DIFF = f"{SOURCE_CODE_ATTRIBUTE_SPACE}upstream_diff_"



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["NeptuneBackend"]

import abc
from typing import (
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from neptune.api.dtos import FileEntry
from neptune.common.exceptions import NeptuneException
from neptune.internal.artifacts.types import ArtifactFileData
from neptune.internal.backends.api_model import (
    ApiExperiment,
    ArtifactAttribute,
    Attribute,
    AttributeType,
    BoolAttribute,
    DatetimeAttribute,
    FileAttribute,
    FloatAttribute,
    FloatSeriesAttribute,
    FloatSeriesValues,
    ImageSeriesValues,
    IntAttribute,
    LeaderboardEntry,
    Project,
    StringAttribute,
    StringSeriesAttribute,
    StringSeriesValues,
    StringSetAttribute,
    Workspace,
)
from neptune.internal.backends.nql import NQLQuery
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import (
    QualifiedName,
    UniqueId,
)
from neptune.internal.operation import Operation
from neptune.internal.operation_processors.operation_storage import OperationStorage
from neptune.internal.utils.git import GitInfo
from neptune.internal.websockets.websockets_factory import WebsocketsFactory


class NeptuneBackend:
    def close(self) -> None:
        """No need for closing implementation"""

    @abc.abstractmethod
    def get_display_address(self) -> str:
        pass

    def verify_feature_available(self, _: str) -> None:
        """
        this method makes sense only for backends interacting with server;
        it makes sure that a feature is supported in the backend version client interacts with
        """

    def websockets_factory(self, project_id: str, run_id: str) -> Optional[WebsocketsFactory]:
        return None

    @abc.abstractmethod
    def get_project(self, project_id: QualifiedName) -> Project:
        pass

    @abc.abstractmethod
    def get_available_projects(
        self, workspace_id: Optional[str] = None, search_term: Optional[str] = None
    ) -> List[Project]:
        pass

    @abc.abstractmethod
    def get_available_workspaces(self) -> List[Workspace]:
        pass

    @abc.abstractmethod
    def create_run(
        self,
        project_id: UniqueId,
        git_info: Optional[GitInfo] = None,
        custom_run_id: Optional[str] = None,
        notebook_id: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ) -> ApiExperiment:
        pass

    @abc.abstractmethod
    def create_model(
        self,
        project_id: UniqueId,
        key: str,
    ) -> ApiExperiment:
        pass

    @abc.abstractmethod
    def create_model_version(
        self,
        project_id: UniqueId,
        model_id: UniqueId,
    ) -> ApiExperiment:
        pass

    @abc.abstractmethod
    def get_metadata_container(
        self,
        container_id: Union[UniqueId, QualifiedName],
        expected_container_type: ContainerType,
    ) -> ApiExperiment:
        pass

    @abc.abstractmethod
    def create_checkpoint(self, notebook_id: str, jupyter_path: str) -> Optional[str]:
        pass

    def ping(self, container_id: str, container_type: ContainerType):
        """Do nothing by default"""

    @abc.abstractmethod
    def execute_operations(
        self,
        container_id: UniqueId,
        container_type: ContainerType,
        operations: List[Operation],
        operation_storage: OperationStorage,
    ) -> Tuple[int, List[NeptuneException]]:
        pass

    @abc.abstractmethod
    def get_attributes(self, container_id: str, container_type: ContainerType) -> List[Attribute]:
        pass

    @abc.abstractmethod
    def download_file(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        destination: Optional[str] = None,
    ):
        pass

    @abc.abstractmethod
    def download_file_set(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        destination: Optional[str] = None,
    ):
        pass

    @abc.abstractmethod
    def get_float_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> FloatAttribute:
        pass

    @abc.abstractmethod
    def get_int_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> IntAttribute:
        pass

    @abc.abstractmethod
    def get_bool_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> BoolAttribute:
        pass

    @abc.abstractmethod
    def get_file_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> FileAttribute:
        pass

    @abc.abstractmethod
    def get_string_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringAttribute:
        pass

    @abc.abstractmethod
    def get_datetime_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> DatetimeAttribute:
        pass

    @abc.abstractmethod
    def get_artifact_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> ArtifactAttribute:
        pass

    @abc.abstractmethod
    def list_artifact_files(self, project_id: str, artifact_hash: str) -> List[ArtifactFileData]:
        pass

    @abc.abstractmethod
    def get_float_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> FloatSeriesAttribute:
        pass

    @abc.abstractmethod
    def get_string_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSeriesAttribute:
        pass

    @abc.abstractmethod
    def get_string_set_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSetAttribute:
        pass

    @abc.abstractmethod
    def download_file_series_by_index(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        index: int,
        destination: str,
    ):
        pass

    @abc.abstractmethod
    def get_image_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        offset: int,
        limit: int,
    ) -> ImageSeriesValues:
        pass

    @abc.abstractmethod
    def get_string_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        offset: int,
        limit: int,
    ) -> StringSeriesValues:
        pass

    @abc.abstractmethod
    def get_float_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        offset: int,
        limit: int,
    ) -> FloatSeriesValues:
        pass

    @abc.abstractmethod
    def get_run_url(self, run_id: str, workspace: str, project_name: str, sys_id: str) -> str:
        pass

    @abc.abstractmethod
    def get_project_url(self, project_id: str, workspace: str, project_name: str) -> str:
        pass

    @abc.abstractmethod
    def get_model_url(self, model_id: str, workspace: str, project_name: str, sys_id: str) -> str:
        pass

    @abc.abstractmethod
    def get_model_version_url(
        self,
        model_version_id: str,
        model_id: str,
        workspace: str,
        project_name: str,
        sys_id: str,
    ) -> str:
        pass

    @abc.abstractmethod
    def fetch_atom_attribute_values(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> List[Tuple[str, AttributeType, Any]]:
        pass

    @abc.abstractmethod
    def search_leaderboard_entries(
        self,
        project_id: UniqueId,
        types: Optional[List[ContainerType]] = None,
        query: Optional[NQLQuery] = None,
        columns: Optional[List[str]] = None,
    ) -> List[LeaderboardEntry]:
        pass

    @abc.abstractmethod
    def list_fileset_files(self, attribute: List[str], container_id: str, path: str) -> List[FileEntry]:
        pass



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["GitRef", "GitRefDisabled"]

from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    NewType,
    Optional,
    TypeVar,
    Union,
)

from neptune.types.atoms.atom import Atom
from neptune.vendor.lib_programname import get_path_executed_script

if TYPE_CHECKING:
    from neptune.types.value_visitor import ValueVisitor

Ret = TypeVar("Ret")
GitRefDisabled = NewType("GitRefDisabled", str)


class WithDisabledMixin:
    DISABLED: GitRefDisabled = GitRefDisabled("DO_NOT_TRACK_GIT_REPOSITORY")
    """Constant that can be used to disable Git repository tracking."""


@dataclass
class GitRef(Atom, WithDisabledMixin):
    """
    Represents Git repository metadata.

    Args:
        repository_path: Path to the repository. If not provided,
            the path to the script that is currently executed is used.
    """

    repository_path: Optional[Union[str, Path]] = get_path_executed_script()

    def accept(self, visitor: "ValueVisitor[Ret]") -> Ret:
        return visitor.visit_git_ref(self)

    def __str__(self) -> str:
        return f"GitRef({self.repository_path})"

    def resolve_path(self) -> Optional[Path]:
        if self.repository_path is None:
            return None
        return Path(self.repository_path).resolve()



__all__ = ["FileEntry"]

import datetime
from dataclasses import dataclass
from typing import Any


@dataclass
class FileEntry:
    name: str
    size: int
    mtime: datetime.datetime
    file_type: str

    @classmethod
    def from_dto(cls, file_dto: Any) -> "FileEntry":
        return cls(name=file_dto.name, size=file_dto.size, mtime=file_dto.mtime, file_type=file_dto.fileType)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["upload_source_code"]

import os
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
)

from neptune.attributes import constants as attr_consts
from neptune.common.storage.storage_utils import normalize_file_name
from neptune.common.utils import is_ipython
from neptune.internal.utils import (
    does_paths_share_common_drive,
    get_absolute_paths,
    get_common_root,
)
from neptune.vendor.lib_programname import (
    empty_path,
    get_path_executed_script,
)

if TYPE_CHECKING:
    from neptune import Run


def upload_source_code(source_files: Optional[List[str]], run: "Run") -> None:
    entrypoint_filepath = get_path_executed_script()

    if not is_ipython() and entrypoint_filepath != empty_path and os.path.isfile(entrypoint_filepath):
        if source_files is None:
            entrypoint = os.path.basename(entrypoint_filepath)
            source_files = str(entrypoint_filepath)
        elif not source_files:
            entrypoint = os.path.basename(entrypoint_filepath)
        else:
            common_root = get_common_root(get_absolute_paths(source_files))
            entrypoint_filepath = os.path.abspath(entrypoint_filepath)

            if common_root is not None and does_paths_share_common_drive([common_root, entrypoint_filepath]):
                entrypoint_filepath = normalize_file_name(os.path.relpath(path=entrypoint_filepath, start=common_root))

            entrypoint = normalize_file_name(entrypoint_filepath)

        run[attr_consts.SOURCE_CODE_ENTRYPOINT_ATTRIBUTE_PATH] = entrypoint

    if source_files is not None:
        run[attr_consts.SOURCE_CODE_FILES_ATTRIBUTE_PATH].upload_files(source_files)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Attribute"]

from typing import (
    TYPE_CHECKING,
    List,
)

from neptune.exceptions import TypeDoesNotSupportAttributeException
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.operation import Operation
from neptune.types.value_copy import ValueCopy

if TYPE_CHECKING:
    from neptune.internal.container_type import ContainerType
    from neptune.metadata_containers import MetadataContainer


class Attribute:
    supports_copy = False

    def __init__(self, container: "MetadataContainer", path: List[str]):
        super().__init__()
        self._container = container
        self._path = path

    def __getattr__(self, attr):
        raise TypeDoesNotSupportAttributeException(type_=type(self), attribute=attr)

    def _enqueue_operation(self, operation: Operation, *, wait: bool):
        self._container._op_processor.enqueue_operation(operation, wait=wait)

    @property
    def _backend(self) -> NeptuneBackend:
        return self._container._backend

    @property
    def _container_id(self) -> str:
        return self._container._id

    @property
    def _container_type(self) -> "ContainerType":
        return self._container.container_type

    def copy(self, value: ValueCopy, wait: bool = False):
        raise Exception(f"{type(self).__name__} doesn't support copying")

    def process_assignment(self, value, wait=False):
        if isinstance(value, ValueCopy):
            return self.copy(value, wait=wait)
        else:
            return self.assign(value, wait=wait)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import re


class CGroupFilesystemReader(object):
    def __init__(self):
        cgroup_memory_dir = self.__cgroup_mount_dir(subsystem="memory")
        self.__memory_usage_file = os.path.join(cgroup_memory_dir, "memory.usage_in_bytes")
        self.__memory_limit_file = os.path.join(cgroup_memory_dir, "memory.limit_in_bytes")

        cgroup_cpu_dir = self.__cgroup_mount_dir(subsystem="cpu")
        self.__cpu_period_file = os.path.join(cgroup_cpu_dir, "cpu.cfs_period_us")
        self.__cpu_quota_file = os.path.join(cgroup_cpu_dir, "cpu.cfs_quota_us")

        cgroup_cpuacct_dir = self.__cgroup_mount_dir(subsystem="cpuacct")
        self.__cpuacct_usage_file = os.path.join(cgroup_cpuacct_dir, "cpuacct.usage")

    def get_memory_usage_in_bytes(self):
        return self.__read_int_file(self.__memory_usage_file)

    def get_memory_limit_in_bytes(self):
        return self.__read_int_file(self.__memory_limit_file)

    def get_cpu_quota_micros(self):
        return self.__read_int_file(self.__cpu_quota_file)

    def get_cpu_period_micros(self):
        return self.__read_int_file(self.__cpu_period_file)

    def get_cpuacct_usage_nanos(self):
        return self.__read_int_file(self.__cpuacct_usage_file)

    def __read_int_file(self, filename):
        with open(filename) as f:
            return int(f.read())

    def __cgroup_mount_dir(self, subsystem):
        """
        :param subsystem: cgroup subsystem like memory, cpu
        :return: directory where given subsystem is mounted
        """
        with open("/proc/mounts", "r") as f:
            for line in f.readlines():
                split_line = re.split(r"\s+", line)
                mount_dir = split_line[1]

                if "cgroup" in mount_dir:
                    dirname = mount_dir.split("/")[-1]
                    subsystems = dirname.split(",")
                    if subsystem in subsystems:
                        return mount_dir

        assert False, 'Mount directory for "{}" subsystem not found'.format(subsystem)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "get_image_content",
    "get_html_content",
    "get_pickle_content",
    "is_numpy_array",
    "is_pil_image",
    "is_matplotlib_figure",
    "is_plotly_figure",
    "is_altair_chart",
    "is_bokeh_figure",
    "is_pandas_dataframe",
]

import base64
import io
import logging
import pickle
import warnings
from io import (
    BytesIO,
    StringIO,
)
from typing import Optional

from packaging import version
from pandas import DataFrame

from neptune.exceptions import PlotlyIncompatibilityException
from neptune.internal.utils.logger import logger

_logger = logging.getLogger(__name__)

try:
    from numpy import array as numpy_array
    from numpy import ndarray as numpy_ndarray
    from numpy import uint8 as numpy_uint8
except ImportError:
    numpy_ndarray = None
    numpy_array = None
    numpy_uint8 = None

try:
    from PIL.Image import Image as PILImage
    from PIL.Image import fromarray as pilimage_fromarray
except ImportError:
    PILImage = None

    def pilimage_fromarray():
        pass


def get_image_content(image) -> Optional[bytes]:
    content = _image_to_bytes(image)

    return content


def get_html_content(chart) -> Optional[str]:
    content = _to_html(chart)

    return content


def get_pickle_content(obj) -> Optional[bytes]:
    content = _export_pickle(obj)

    return content


def _image_to_bytes(image) -> bytes:
    if image is None:
        raise ValueError("image is None")

    elif is_numpy_array(image):
        return _get_numpy_as_image(image)

    elif is_pil_image(image):
        return _get_pil_image_data(image)

    elif is_matplotlib_figure(image):
        return _get_figure_image_data(image)

    elif _is_torch_tensor(image):
        return _get_numpy_as_image(image.detach().numpy())

    elif _is_tensorflow_tensor(image):
        return _get_numpy_as_image(image.numpy())

    raise TypeError("image is {}".format(type(image)))


def _to_html(chart) -> str:
    if _is_matplotlib_pyplot(chart):
        chart = chart.gcf()

    if is_matplotlib_figure(chart):
        try:
            chart = _matplotlib_to_plotly(chart)
            return _export_plotly_figure(chart)
        except ImportError:
            logger.warning("Plotly not installed. Logging plot as an image.")
            return _image_content_to_html(_get_figure_image_data(chart))
        except UserWarning:
            logger.warning(
                "Couldn't convert Matplotlib plot to interactive Plotly plot. Logging plot as an image instead."
            )
            return _image_content_to_html(_get_figure_image_data(chart))

    elif is_pandas_dataframe(chart):
        return _export_pandas_dataframe_to_html(chart)

    elif is_plotly_figure(chart):
        return _export_plotly_figure(chart)

    elif is_altair_chart(chart):
        return _export_altair_chart(chart)

    elif is_bokeh_figure(chart):
        return _export_bokeh_figure(chart)

    else:
        raise ValueError("Currently supported are matplotlib, plotly, altair, and bokeh figures")


def _matplotlib_to_plotly(chart):
    import matplotlib
    import plotly

    # When Plotly cannot accurately convert a matplotlib plot, it emits a warning.
    # Then we want to fallback on logging the plot as an image.
    #
    # E.g. when trying to convert a Seaborn confusion matrix or a hist2d, it emits a UserWarning with message
    # "Dang! That path collection is out of this world. I totally don't know what to do with it yet!
    # Plotly can only import path collections linked to 'data' coordinates"
    #
    # Plotly is not compatible with the latest matplotlib (3.5.0+)
    # due to fact that mpl_to_plotly uses deprecated matplotlib functionalities
    plotly_version = plotly.__version__
    matplotlib_version = matplotlib.__version__
    if version.parse(matplotlib_version) >= version.parse("3.3.0") and version.parse(plotly_version) < version.parse(
        "5.0.0"
    ):
        raise PlotlyIncompatibilityException(
            matplotlib_version,
            plotly_version,
            "Downgrade matplotlib to version 3.2, upgrade plotly to 5.0+, or upload the chart as a static "
            "image: run['chart'].upload(File.as_image(plotly_chart)). "
            "For details, see https://github.com/plotly/plotly.py/issues/1568.",
        )

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "error",
            category=UserWarning,
            message=".*Plotly can only import path collections linked to 'data' coordinates.*",
        )
        try:
            chart = plotly.tools.mpl_to_plotly(chart)
        except AttributeError as e:
            if "'PathCollection' object has no attribute 'get_offset_position'" in str(e):
                raise PlotlyIncompatibilityException(
                    matplotlib_version,
                    plotly_version,
                    "Due to plotly using some deprecated matplotlib methods, we recommend downgrading matplotlib"
                    " to version 3.4. See https://github.com/plotly/plotly.py/issues/3624 for details.",
                ) from e
            raise e

    return chart


def _image_content_to_html(content: bytes) -> str:
    str_equivalent_image = base64.b64encode(content).decode()
    return "<img src='data:image/png;base64," + str_equivalent_image + "'/>"


def _get_numpy_as_image(array):
    array = array.copy()  # prevent original array from modifying

    data_range_warnings = []
    array_min = array.min()
    array_max = array.max()
    if array_min < 0:
        data_range_warnings.append(f"the smallest value in the array is {array_min}")
    if array_max > 1:
        data_range_warnings.append(f"the largest value in the array is {array_max}")
    if data_range_warnings:
        data_range_warning_message = (" and ".join(data_range_warnings) + ".").capitalize()
        logger.warning(
            "%s To be interpreted as colors correctly values in the array need to be in the [0, 1] range.",
            data_range_warning_message,
        )
    array *= 255
    shape = array.shape
    if len(shape) == 2:
        return _get_pil_image_data(pilimage_fromarray(array.astype(numpy_uint8)))
    if len(shape) == 3:
        if shape[2] == 1:
            array2d = numpy_array([[col[0] for col in row] for row in array])
            return _get_pil_image_data(pilimage_fromarray(array2d.astype(numpy_uint8)))
        if shape[2] in (3, 4):
            return _get_pil_image_data(pilimage_fromarray(array.astype(numpy_uint8)))
    raise ValueError(
        "Incorrect size of numpy.ndarray. Should be 2-dimensional or"
        "3-dimensional with 3rd dimension of size 1, 3 or 4."
    )


def _get_pil_image_data(image: PILImage) -> bytes:
    with io.BytesIO() as image_buffer:
        image.save(image_buffer, format="PNG")
        return image_buffer.getvalue()


def _get_figure_image_data(figure) -> bytes:
    with io.BytesIO() as image_buffer:
        figure.savefig(image_buffer, format="png", bbox_inches="tight")
        return image_buffer.getvalue()


def _is_torch_tensor(image):
    return (
        image.__class__.__module__.startswith("torch")
        and image.__class__.__name__ == "Tensor"
        and hasattr(image, "numpy")
    )


def _is_tensorflow_tensor(image):
    return (
        image.__class__.__module__.startswith("tensorflow.")
        and "Tensor" in image.__class__.__name__
        and hasattr(image, "numpy")
    )


def _is_matplotlib_pyplot(chart):
    return chart.__class__.__module__.startswith("matplotlib.pyplot")


def is_numpy_array(image) -> bool:
    return numpy_ndarray is not None and isinstance(image, numpy_ndarray)


def is_pil_image(image) -> bool:
    return PILImage is not None and isinstance(image, PILImage)


def is_matplotlib_figure(image):
    return image.__class__.__module__.startswith("matplotlib.") and image.__class__.__name__ == "Figure"


def is_plotly_figure(chart):
    return chart.__class__.__module__.startswith("plotly.") and chart.__class__.__name__ == "Figure"


def is_altair_chart(chart):
    return chart.__class__.__module__.startswith("altair.") and "Chart" in chart.__class__.__name__


def is_bokeh_figure(chart):
    return chart.__class__.__module__.startswith("bokeh.") and chart.__class__.__name__.lower() == "figure"


def is_pandas_dataframe(table):
    return isinstance(table, DataFrame)


def _export_pandas_dataframe_to_html(table):
    buffer = StringIO(table.to_html())
    buffer.seek(0)
    return buffer.getvalue()


def _export_plotly_figure(image):
    buffer = StringIO()
    image.write_html(buffer)
    buffer.seek(0)
    return buffer.getvalue()


def _export_altair_chart(chart):
    buffer = StringIO()
    chart.save(buffer, format="html")
    buffer.seek(0)
    return buffer.getvalue()


def _export_bokeh_figure(chart):
    from bokeh.embed import file_html
    from bokeh.resources import CDN

    html = file_html(chart, CDN)
    buffer = StringIO(html)
    buffer.seek(0)
    return buffer.getvalue()


def _export_pickle(obj):
    buffer = BytesIO()
    pickle.dump(obj, buffer)
    buffer.seek(0)
    return buffer.getvalue()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["FileSet"]

import os
from typing import (
    Iterable,
    List,
    Optional,
    Union,
)

from neptune.api.dtos import FileEntry
from neptune.attributes.attribute import Attribute
from neptune.internal.operation import (
    DeleteFiles,
    UploadFileSet,
)
from neptune.internal.utils import (
    verify_collection_type,
    verify_type,
)
from neptune.types.file_set import FileSet as FileSetVal


class FileSet(Attribute):
    def assign(self, value: Union[FileSetVal, str, Iterable[str]], *, wait: bool = False) -> None:
        verify_type("value", value, (FileSetVal, str, Iterable))
        if isinstance(value, FileSetVal):
            value = value.file_globs
        elif isinstance(value, str):
            value = [value]
        else:
            verify_collection_type("value", value, str)
        self._enqueue_upload_operation(value, reset=True, wait=wait)

    def upload_files(self, globs: Union[str, Iterable[str]], *, wait: bool = False) -> None:
        if isinstance(globs, str):
            globs = [globs]
        else:
            verify_collection_type("globs", globs, str)
        self._enqueue_upload_operation(globs, reset=False, wait=wait)

    def delete_files(self, paths: Union[str, Iterable[str]], *, wait: bool = False) -> None:
        if isinstance(paths, str):
            paths = [paths]
        else:
            verify_collection_type("paths", paths, str)
        with self._container.lock():
            self._enqueue_operation(DeleteFiles(self._path, set(paths)), wait=wait)

    def _enqueue_upload_operation(self, globs: Iterable[str], *, reset: bool, wait: bool):
        with self._container.lock():
            abs_file_globs = list(os.path.abspath(file_glob) for file_glob in globs)
            self._enqueue_operation(UploadFileSet(self._path, abs_file_globs, reset=reset), wait=wait)

    def download(self, destination: Optional[str] = None) -> None:
        verify_type("destination", destination, (str, type(None)))
        self._backend.download_file_set(self._container_id, self._container_type, self._path, destination)

    def list_fileset_files(self, path: Optional[str] = None) -> List[FileEntry]:
        path = path or ""
        return self._backend.list_fileset_files(self._path, self._container_id, path)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import dataclasses
import io
import math
import os
import tarfile
from typing import (
    Any,
    Generator,
    Optional,
)

from neptune.common.backends.api_model import MultipartConfig
from neptune.common.exceptions import (
    InternalClientError,
    UploadedFileChanged,
)


@dataclasses.dataclass
class FileChunk:
    data: bytes
    start: int
    end: int


class FileChunker:
    def __init__(self, filename: Optional[str], fobj, total_size, multipart_config: MultipartConfig):
        self._filename: Optional[str] = filename
        self._fobj = fobj
        self._total_size = total_size
        self._min_chunk_size = multipart_config.min_chunk_size
        self._max_chunk_size = multipart_config.max_chunk_size
        self._max_chunk_count = multipart_config.max_chunk_count

    def _get_chunk_size(self) -> int:
        if self._total_size > self._max_chunk_count * self._max_chunk_size:
            # can't fit it
            max_size = self._max_chunk_count * self._max_chunk_size
            raise InternalClientError(
                f"File {self._filename or 'stream'} is too big to upload:"
                f" {self._total_size} bytes exceeds max size {max_size}"
            )
        if self._total_size <= self._max_chunk_count * self._min_chunk_size:
            # can be done as minimal size chunks -- go for it!
            return self._min_chunk_size
        else:
            # need larger chunks -- split more or less equally
            return math.ceil(self._total_size / self._max_chunk_count)

    def generate(self) -> Generator[FileChunk, Any, None]:
        chunk_size = self._get_chunk_size()
        last_offset = 0
        last_change: Optional = os.stat(self._filename).st_mtime if self._filename else None
        while last_offset < self._total_size:
            chunk = self._fobj.read(chunk_size)
            if chunk:
                if last_change and last_change < os.stat(self._filename).st_mtime:
                    raise UploadedFileChanged(self._filename)
                if isinstance(chunk, str):
                    chunk = chunk.encode("utf-8")
                new_offset = last_offset + len(chunk)
                yield FileChunk(data=chunk, start=last_offset, end=new_offset)
                last_offset = new_offset


def compress_to_tar_gz_in_memory(upload_entries) -> bytes:
    f = io.BytesIO(b"")

    with tarfile.TarFile.open(fileobj=f, mode="w|gz", dereference=True) as archive:
        for entry in upload_entries:
            archive.add(name=entry.source, arcname=entry.target_path, recursive=True)

    f.seek(0)
    data = f.read()
    return data



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["WebsocketSignalsBackgroundJob"]

import json
import logging
import threading
from json.decoder import JSONDecodeError
from typing import (
    TYPE_CHECKING,
    Optional,
)

from websocket import WebSocketConnectionClosedException

from neptune.attributes.constants import (
    SIGNAL_TYPE_ABORT,
    SIGNAL_TYPE_STOP,
    SYSTEM_FAILED_ATTRIBUTE_PATH,
)
from neptune.common.websockets.reconnecting_websocket import ReconnectingWebsocket
from neptune.internal.background_job import BackgroundJob
from neptune.internal.threading.daemon import Daemon
from neptune.internal.utils import process_killer
from neptune.internal.utils.logger import logger
from neptune.internal.websockets.websockets_factory import WebsocketsFactory

if TYPE_CHECKING:
    from neptune.metadata_containers import MetadataContainer

_logger = logging.getLogger(__name__)


class WebsocketSignalsBackgroundJob(BackgroundJob):
    def __init__(self, ws_factory: WebsocketsFactory):
        self._ws_factory = ws_factory
        self._thread: "Optional[WebsocketSignalsBackgroundJob._ListenerThread]" = None
        self._started = False

    def start(self, container: "MetadataContainer"):
        self._thread = self._ListenerThread(container, self._ws_factory.create())
        self._thread.start()
        self._started = True

    def stop(self):
        if not self._started:
            return
        self._thread.interrupt()
        self._thread.shutdown_ws_client()

    def pause(self):
        pass

    def resume(self):
        pass

    def join(self, seconds: Optional[float] = None):
        if not self._started or threading.get_ident() == self._thread.ident:
            return
        self._thread.join(seconds)
        # Just in case. There is possible race condition when connection can be reestablished after being shutdown.
        self._thread.shutdown_ws_client()

    class _ListenerThread(Daemon):
        def __init__(self, container: "MetadataContainer", ws_client: ReconnectingWebsocket):
            super().__init__(sleep_time=0, name="NeptuneWebhooks")
            self._container = container
            self._ws_client = ws_client

        def work(self) -> None:
            try:
                raw_message = self._ws_client.recv()
                if raw_message is None or self._is_heartbeat(raw_message):
                    return
                else:
                    self._handler_message(raw_message)
            except WebSocketConnectionClosedException:
                pass

        def _handler_message(self, msg: str):
            try:
                json_msg = json.loads(msg)
                msg_type = json_msg.get("type")
                msg_body = json_msg.get("body")
                if not msg_type:
                    logger.error("Malformed websocket signal: missing type")
                    return
                if not isinstance(msg_type, str):
                    logger.error("Malformed websocket signal: type is %s, should be str", type(msg_type))
                    return
                if msg_type.lower() == SIGNAL_TYPE_STOP:
                    self._handle_stop(msg_body)
                elif msg_type.lower() == SIGNAL_TYPE_ABORT:
                    self._handle_abort(msg_body)
            except JSONDecodeError as ex:
                logger.error("Malformed websocket signal: %s, message: %s", ex, msg)

        def _handle_stop(self, msg_body):
            msg_body = msg_body or dict()
            if not isinstance(msg_body, dict):
                logger.error("Malformed websocket signal: body is %s, should be dict", type(msg_body))
                return
            run_id = self._container["sys/id"].fetch()
            logger.error("Run %s received stop signal. Exiting", run_id)
            seconds = msg_body.get("seconds")
            self._container.stop(seconds=seconds)
            process_killer.kill_me()

        def _handle_abort(self, msg_body):
            msg_body = msg_body or dict()
            if not isinstance(msg_body, dict):
                logger.error("Malformed websocket signal: body is %s, should be dict", type(msg_body))
                return
            run_id = self._container["sys/id"].fetch()
            logger.error("Run %s received abort signal. Exiting", run_id)
            seconds = msg_body.get("seconds")
            self._container[SYSTEM_FAILED_ATTRIBUTE_PATH] = True
            self._container.stop(seconds=seconds)
            process_killer.kill_me()

        def shutdown_ws_client(self):
            self._ws_client.shutdown()

        @staticmethod
        def _is_heartbeat(message: str):
            return message.strip() == ""



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "Project",
    "Workspace",
    "ApiExperiment",
    "OptionalFeatures",
    "VersionInfo",
    "ClientConfig",
    "AttributeType",
    "Attribute",
    "AttributeWithProperties",
    "LeaderboardEntry",
    "StringPointValue",
    "ImageSeriesValues",
    "StringSeriesValues",
    "FloatPointValue",
    "FloatSeriesValues",
    "FloatAttribute",
    "IntAttribute",
    "BoolAttribute",
    "FileAttribute",
    "StringAttribute",
    "DatetimeAttribute",
    "ArtifactAttribute",
    "ArtifactModel",
    "FloatSeriesAttribute",
    "StringSeriesAttribute",
    "StringSetAttribute",
]

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import (
    Any,
    FrozenSet,
    List,
    Optional,
    Set,
)

from packaging import version

from neptune.common.backends.api_model import MultipartConfig
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import (
    SysId,
    UniqueId,
)


@dataclass
class Project:
    id: UniqueId
    name: str
    workspace: str
    sys_id: SysId


@dataclass
class Workspace:
    id: UniqueId
    name: str


@dataclass
class ApiExperiment:
    id: UniqueId
    type: ContainerType
    sys_id: SysId
    workspace: str
    project_name: str
    trashed: bool = False

    @classmethod
    def from_experiment(cls, response_exp):
        return cls(
            id=response_exp.id,
            type=ContainerType.from_api(response_exp.type),
            sys_id=response_exp.shortId,
            workspace=response_exp.organizationName,
            project_name=response_exp.projectName,
            trashed=response_exp.trashed,
        )


class OptionalFeatures:
    VERSION_INFO = "version_info"
    ARTIFACTS = "artifacts"
    ARTIFACTS_HASH_EXCLUDE_METADATA = "artifacts_hash_exclude_metadata"
    ARTIFACTS_EXCLUDE_DIRECTORY_FILES = "artifact_exclude_directory_files"
    MULTIPART_UPLOAD = "multipart_upload"


@dataclass(frozen=True)
class VersionInfo:
    min_recommended: Optional[version.Version]
    min_compatible: Optional[version.Version]
    max_compatible: Optional[version.Version]

    @staticmethod
    def build(
        min_recommended: Optional[str],
        min_compatible: Optional[str],
        max_compatible: Optional[str],
    ) -> "VersionInfo":
        return VersionInfo(
            min_recommended=version.parse(min_recommended) if min_recommended else None,
            min_compatible=version.parse(min_compatible) if min_compatible else None,
            max_compatible=version.parse(max_compatible) if max_compatible else None,
        )


@dataclass(frozen=True)
class ClientConfig:
    api_url: str
    display_url: str
    _missing_features: FrozenSet[str]
    version_info: VersionInfo
    multipart_config: MultipartConfig

    def has_feature(self, feature_name: str) -> bool:
        return feature_name not in self._missing_features

    @staticmethod
    def from_api_response(config) -> "ClientConfig":
        missing_features = []

        version_info_obj = getattr(config, "pyLibVersions", None)
        if version_info_obj is None:
            missing_features.append(OptionalFeatures.VERSION_INFO)
            min_recommended = min_compatible = max_compatible = None
        else:
            min_recommended = getattr(version_info_obj, "minRecommendedVersion", None)
            min_compatible = getattr(version_info_obj, "minCompatibleVersion", None)
            max_compatible = getattr(version_info_obj, "maxCompatibleVersion", None)

        multipart_upload_config_obj = getattr(config, "multiPartUpload", None)
        has_multipart_upload = getattr(multipart_upload_config_obj, "enabled", False)
        if not has_multipart_upload:
            missing_features.append(OptionalFeatures.MULTIPART_UPLOAD)
            multipart_upload_config = None
        else:
            min_chunk_size = getattr(multipart_upload_config_obj, "minChunkSize")
            max_chunk_size = getattr(multipart_upload_config_obj, "maxChunkSize")
            max_chunk_count = getattr(multipart_upload_config_obj, "maxChunkCount")
            max_single_part_size = getattr(multipart_upload_config_obj, "maxSinglePartSize")
            multipart_upload_config = MultipartConfig(
                min_chunk_size, max_chunk_size, max_chunk_count, max_single_part_size
            )

        artifacts_config_obj = getattr(config, "artifacts", None)
        has_artifacts = getattr(artifacts_config_obj, "enabled", False)
        if not has_artifacts:
            missing_features.append(OptionalFeatures.ARTIFACTS)

        artifacts_api_version = getattr(artifacts_config_obj, "apiVersion", 1)
        if artifacts_api_version == 1:
            missing_features.append(OptionalFeatures.ARTIFACTS_HASH_EXCLUDE_METADATA)
            missing_features.append(OptionalFeatures.ARTIFACTS_EXCLUDE_DIRECTORY_FILES)

        return ClientConfig(
            api_url=config.apiUrl,
            display_url=config.applicationUrl,
            _missing_features=frozenset(missing_features),
            version_info=VersionInfo.build(min_recommended, min_compatible, max_compatible),
            multipart_config=multipart_upload_config,
        )


class AttributeType(Enum):
    FLOAT = "float"
    INT = "int"
    BOOL = "bool"
    STRING = "string"
    DATETIME = "datetime"
    FILE = "file"
    FILE_SET = "fileSet"
    FLOAT_SERIES = "floatSeries"
    STRING_SERIES = "stringSeries"
    IMAGE_SERIES = "imageSeries"
    STRING_SET = "stringSet"
    GIT_REF = "gitRef"
    RUN_STATE = "experimentState"
    NOTEBOOK_REF = "notebookRef"
    ARTIFACT = "artifact"


@dataclass
class Attribute:
    path: str
    type: AttributeType


@dataclass
class AttributeWithProperties:
    path: str
    type: AttributeType
    properties: Any


@dataclass
class LeaderboardEntry:
    id: str
    attributes: List[AttributeWithProperties]


@dataclass
class StringPointValue:
    timestampMillis: int
    step: float
    value: str


@dataclass
class ImageSeriesValues:
    totalItemCount: int


@dataclass
class StringSeriesValues:
    totalItemCount: int
    values: List[StringPointValue]


@dataclass
class FloatPointValue:
    timestampMillis: int
    step: float
    value: float


@dataclass
class FloatSeriesValues:
    totalItemCount: int
    values: List[FloatPointValue]


@dataclass
class FloatAttribute:
    value: float


@dataclass
class IntAttribute:
    value: int


@dataclass
class BoolAttribute:
    value: bool


@dataclass
class FileAttribute:
    name: str
    ext: str
    size: int


@dataclass
class StringAttribute:
    value: str


@dataclass
class DatetimeAttribute:
    value: datetime


@dataclass
class ArtifactAttribute:
    hash: str


@dataclass
class ArtifactModel:
    received_metadata: bool
    hash: str
    size: int


@dataclass
class FloatSeriesAttribute:
    last: Optional[float]


@dataclass
class StringSeriesAttribute:
    last: Optional[str]


@dataclass
class StringSetAttribute:
    values: Set[str]



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["ModelVersion"]

import os
from typing import Optional

from neptune.attributes.constants import (
    SYSTEM_NAME_ATTRIBUTE_PATH,
    SYSTEM_STAGE_ATTRIBUTE_PATH,
)
from neptune.common.exceptions import NeptuneException
from neptune.envs import CONNECTION_MODE
from neptune.exceptions import (
    InactiveModelVersionException,
    NeedExistingModelVersionForReadOnlyMode,
    NeptuneMissingRequiredInitParameter,
    NeptuneOfflineModeChangeStageException,
)
from neptune.internal.backends.api_model import ApiExperiment
from neptune.internal.backgroud_job_list import BackgroundJobList
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import QualifiedName
from neptune.internal.init.parameters import (
    ASYNC_LAG_THRESHOLD,
    ASYNC_NO_PROGRESS_THRESHOLD,
    DEFAULT_FLUSH_PERIOD,
    DEFAULT_NAME,
    OFFLINE_PROJECT_QUALIFIED_NAME,
)
from neptune.internal.operation_processors.offline_operation_processor import OfflineOperationProcessor
from neptune.internal.state import ContainerState
from neptune.internal.utils import verify_type
from neptune.internal.utils.ping_background_job import PingBackgroundJob
from neptune.metadata_containers import MetadataContainer
from neptune.metadata_containers.abstract import NeptuneObjectCallback
from neptune.types.mode import Mode
from neptune.types.model_version_stage import ModelVersionStage


class ModelVersion(MetadataContainer):
    """Class for managing a version of a neptune.ai model and retrieving information from it."""

    container_type = ContainerType.MODEL_VERSION

    def __init__(
        self,
        with_id: Optional[str] = None,
        *,
        name: Optional[str] = None,
        model: Optional[str] = None,
        project: Optional[str] = None,
        api_token: Optional[str] = None,
        mode: Optional[str] = None,
        flush_period: float = DEFAULT_FLUSH_PERIOD,
        proxies: Optional[dict] = None,
        async_lag_callback: Optional[NeptuneObjectCallback] = None,
        async_lag_threshold: float = ASYNC_LAG_THRESHOLD,
        async_no_progress_callback: Optional[NeptuneObjectCallback] = None,
        async_no_progress_threshold: float = ASYNC_NO_PROGRESS_THRESHOLD,
    ) -> None:
        """Initializes a ModelVersion object from an existing or new model version.

        Before creating model versions, you must first register a model by creating a Model object.

        A ModelVersion object is suitable for storing model metadata that is version-specific. It does not track
        background metrics or logs automatically, but you can assign metadata to the model version just like you can
        for runs. You can use the parent Model object to store metadata that is common to all versions of the model.
        To learn more about model registry, see the docs: https://docs.neptune.ai/model_registry/overview/

        To manage the stage of a model version, use its `change_stage()` method or use the menu in the web app.

        You can also use the ModelVersion object as a context manager (see examples).

        Args:
            with_id: The Neptune identifier of an existing model version to resume, such as "CLS-PRE-3".
                The identifier is stored in the object's "sys/id" field.
                If omitted or None is passed, a new model version is created.
            name: A custom name for the model version. Defaults to "Untitled".
            model: Identifier of the model for which the new version should be created.
                Required when creating a new model version. The identifier is stored in the model's "sys/id" field.
            project: Name of a project in the form `workspace-name/project-name`.
                If None, the value of the NEPTUNE_PROJECT environment variable is used.
            api_token: User's API token.
                If None (default), the value of the NEPTUNE_API_TOKEN environment variable is used.
                Note: To keep your API token secure, save it to the NEPTUNE_API_TOKEN environment variable rather than
                placing it in plain text in the source code.
            mode: Connection mode in which the tracking will work.
                If None (default), the value of the NEPTUNE_MODE environment variable is used.
                If no value was set for the environment variable, "async" is used by default.
                Possible values are `async`, `sync`, `offline`, `read-only`, and `debug`.
            flush_period: In the asynchronous (default) connection mode, how often disk flushing is triggered
                (in seconds).
            proxies: Argument passed to HTTP calls made via the Requests library, as dictionary of strings.
                For more information about proxies, see the Requests documentation.
            async_lag_callback: Custom callback which is called if the lag between a queued operation and its
                synchronization with the server exceeds the duration defined by `async_lag_threshold`. The callback
                should take a ModelVersion object as the argument and can contain any custom code, such as calling
                `stop()` on the object.
                Note: Instead of using this argument, you can use Neptune's default callback by setting the
                `NEPTUNE_ENABLE_DEFAULT_ASYNC_LAG_CALLBACK` environment variable to `TRUE`.
            async_lag_threshold: In seconds, duration between the queueing and synchronization of an operation.
                If a lag callback (default callback enabled via environment variable or custom callback passed to the
                `async_lag_callback` argument) is enabled, the callback is called when this duration is exceeded.
            async_no_progress_callback: Custom callback which is called if there has been no synchronization progress
                whatsoever for the duration defined by `async_no_progress_threshold`. The callback should take a
                ModelVersion object as the argument and can contain any custom code, such as calling `stop()` on the
                object.
                Note: Instead of using this argument, you can use Neptune's default callback by setting the
                `NEPTUNE_ENABLE_DEFAULT_ASYNC_NO_PROGRESS_CALLBACK` environment variable to `TRUE`.
            async_no_progress_threshold: In seconds, for how long there has been no synchronization progress since the
                object was initialized. If a no-progress callback (default callback enabled via environment variable or
                custom callback passed to the `async_no_progress_callback` argument) is enabled, the callback is called
                when this duration is exceeded.

        Returns:
            ModelVersion object that is used to manage the model version and log metadata to it.

        Examples:

            >>> import neptune

            Creating a new model version:

            >>> # Create a new model version for a model with identifier "CLS-PRE"
            ... model_version = neptune.init_model_version(model="CLS-PRE")
            >>> model_version["your/structure"] = some_metadata

            >>> # You can provide the project parameter as an environment variable
            ... # or directly in the init_model_version() function:
            ... model_version = neptune.init_model_version(
            ...    model="CLS-PRE",
            ...    project="ml-team/classification",
            ... )

            >>> # Or initialize with the constructor:
            ... model_version = ModelVersion(model="CLS-PRE")

            Connecting to an existing model version:

            >>> # Initialize an existing model version with identifier "CLS-PRE-12"
            ... model_version = neptune.init_model_version(with_id="CLS-PRE-12")

            >>> # To prevent modifications when connecting to an existing model version,
            ... # you can connect in read-only mode:
            ... model_version = neptune.init_model(with_id="CLS-PRE-12", mode="read-only")

            Using the ModelVersion object as context manager:

            >>> with ModelVersion(model="CLS-PRE") as model_version:
            ...     model_version["metadata"] = some_metadata

        For more, see the docs:
            Initializing a model version:
                https://docs.neptune.ai/api/neptune#init_model_version
            ModelVersion class reference:
                https://docs.neptune.ai/api/model_version/
        """
        verify_type("with_id", with_id, (str, type(None)))
        verify_type("name", name, (str, type(None)))
        verify_type("model", model, (str, type(None)))
        verify_type("project", project, (str, type(None)))
        verify_type("mode", mode, (str, type(None)))

        self._model: Optional[str] = model
        self._with_id: Optional[str] = with_id
        self._name: Optional[str] = DEFAULT_NAME if model is None and name is None else name

        # make mode proper Enum instead of string
        mode = Mode(mode or os.getenv(CONNECTION_MODE) or Mode.ASYNC.value)

        if mode == Mode.OFFLINE:
            raise NeptuneException("ModelVersion can't be initialized in OFFLINE mode")

        if mode == Mode.DEBUG:
            project = OFFLINE_PROJECT_QUALIFIED_NAME

        super().__init__(
            project=project,
            api_token=api_token,
            mode=mode,
            flush_period=flush_period,
            proxies=proxies,
            async_lag_callback=async_lag_callback,
            async_lag_threshold=async_lag_threshold,
            async_no_progress_callback=async_no_progress_callback,
            async_no_progress_threshold=async_no_progress_threshold,
        )

    def _get_or_create_api_object(self) -> ApiExperiment:
        project_workspace = self._project_api_object.workspace
        project_name = self._project_api_object.name
        project_qualified_name = f"{project_workspace}/{project_name}"

        if self._with_id is not None:
            # with_id (resume existing model_version) has priority over model (creating a new model_version)
            return self._backend.get_metadata_container(
                container_id=QualifiedName(project_qualified_name + "/" + self._with_id),
                expected_container_type=self.container_type,
            )
        elif self._model is not None:
            if self._mode == Mode.READ_ONLY:
                raise NeedExistingModelVersionForReadOnlyMode()

            api_model = self._backend.get_metadata_container(
                container_id=QualifiedName(project_qualified_name + "/" + self._model),
                expected_container_type=ContainerType.MODEL,
            )
            return self._backend.create_model_version(project_id=self._project_api_object.id, model_id=api_model.id)
        else:
            raise NeptuneMissingRequiredInitParameter(
                parameter_name="model",
                called_function="init_model_version",
            )

    def _prepare_background_jobs(self) -> BackgroundJobList:
        return BackgroundJobList([PingBackgroundJob()])

    def _write_initial_attributes(self):
        if self._name is not None:
            self[SYSTEM_NAME_ATTRIBUTE_PATH] = self._name

    def _raise_if_stopped(self):
        if self._state == ContainerState.STOPPED:
            raise InactiveModelVersionException(label=self._sys_id)

    def get_url(self) -> str:
        """Returns the URL that can be accessed within the browser"""
        return self._backend.get_model_version_url(
            model_version_id=self._id,
            workspace=self._workspace,
            project_name=self._project_name,
            sys_id=self._sys_id,
            model_id=self["sys/model_id"].fetch(),
        )

    def change_stage(self, stage: str) -> None:
        """Changes the stage of the model version.

        This method is always synchronous, which means that Neptune will wait for all other calls to reach the Neptune
            servers before executing it.
        Args:
            stage: The new stage of the model version.
                Possible values are `none`, `staging`, `production`, and `archived`.
        Examples:
            >>> import neptune
            >>> model_version = neptune.init_model_version(with_id="CLS-TREE-3")
            >>> # If the model is good enough, promote it to the staging
            ... val_acc = model_version["validation/metrics/acc"].fetch()
            >>> if val_acc >= ACC_THRESHOLD:
            ...     model_version.change_stage("staging")

        Learn more about stage management in the docs:
            https://docs.neptune.ai/model_registry/managing_stage/
        API reference:
            https://docs.neptune.ai/api/model_version/#change_stage
        """
        mapped_stage = ModelVersionStage(stage)

        if isinstance(self._op_processor, OfflineOperationProcessor):
            raise NeptuneOfflineModeChangeStageException()

        self.wait()

        with self.lock():
            attr = self.get_attribute(SYSTEM_STAGE_ATTRIBUTE_PATH)
            # We are sure that such attribute exists, because
            # SYSTEM_STAGE_ATTRIBUTE_PATH is set by default on ModelVersion creation
            assert attr is not None, f"No {SYSTEM_STAGE_ATTRIBUTE_PATH} found in model version"
            attr.process_assignment(
                value=mapped_stage.value,
                wait=True,
            )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["FileSeries"]

import io
import os
import pathlib
from typing import (
    Iterable,
    List,
    Optional,
)

from PIL import (
    Image,
    UnidentifiedImageError,
)

from neptune.attributes.series.series import Series
from neptune.exceptions import (
    FileNotFound,
    OperationNotSupported,
)
from neptune.internal.operation import (
    ClearImageLog,
    ImageValue,
    LogImages,
    Operation,
)
from neptune.internal.types.file_types import FileType
from neptune.internal.utils import base64_encode
from neptune.internal.utils.limits import image_size_exceeds_limit_for_logging
from neptune.types import File
from neptune.types.series.file_series import FileSeries as FileSeriesVal

Val = FileSeriesVal
Data = File
LogOperation = LogImages


class FileSeries(Series[Val, Data, LogOperation], max_batch_size=1, operation_cls=LogOperation):
    @classmethod
    def _map_series_val(cls, value: Val) -> List[ImageValue]:
        return [
            ImageValue(
                data=cls._get_base64_image_content(val),
                name=value.name,
                description=value.description,
            )
            for val in value.values
        ]

    def _get_clear_operation(self) -> Operation:
        return ClearImageLog(self._path)

    def _data_to_value(self, values: Iterable, **kwargs) -> Val:
        return FileSeriesVal(values, **kwargs)

    def _is_value_type(self, value) -> bool:
        return isinstance(value, FileSeriesVal)

    @staticmethod
    def _get_base64_image_content(file: File) -> str:
        if file.file_type is FileType.LOCAL_FILE:
            if not os.path.exists(file.path):
                raise FileNotFound(file.path)
            with open(file.path, "rb") as image_file:
                file_content = File.from_stream(image_file).content
        else:
            file_content = file.content

        try:
            with Image.open(io.BytesIO(file_content)):
                ...
        except UnidentifiedImageError:
            raise OperationNotSupported(
                "FileSeries supports only image files for now. Other file types will be implemented in future."
            )

        if image_size_exceeds_limit_for_logging(len(file_content)):
            file_content = b""

        return base64_encode(file_content)

    def download(self, destination: Optional[str]):
        target_dir = self._get_destination(destination)
        item_count = self._backend.get_image_series_values(
            self._container_id, self._container_type, self._path, 0, 1
        ).totalItemCount
        for i in range(0, item_count):
            self._backend.download_file_series_by_index(
                self._container_id, self._container_type, self._path, i, target_dir
            )

    def download_last(self, destination: Optional[str]):
        target_dir = self._get_destination(destination)
        item_count = self._backend.get_image_series_values(
            self._container_id, self._container_type, self._path, 0, 1
        ).totalItemCount
        if item_count > 0:
            self._backend.download_file_series_by_index(
                self._container_id,
                self._container_type,
                self._path,
                item_count - 1,
                target_dir,
            )
        else:
            raise ValueError("Unable to download last file - series is empty")

    def _get_destination(self, destination: Optional[str]):
        target_dir = destination
        if destination is None:
            target_dir = os.path.join("neptune", self._path[-1])
        pathlib.Path(os.path.abspath(target_dir)).mkdir(parents=True, exist_ok=True)
        return target_dir



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "Attribute",
    "NamespaceAttr",
    "NamespaceBuilder",
    "InactiveRunException",
    "MetadataInconsistency",
    "NeptunePossibleLegacyUsageException",
    "Handler",
    "RunState",
    "Run",
    "Boolean",
    "Integer",
    "Datetime",
    "Float",
    "String",
    "Namespace",
    "Value",
]

# backwards compatibility
from neptune.attributes.attribute import Attribute
from neptune.attributes.namespace import Namespace as NamespaceAttr
from neptune.attributes.namespace import NamespaceBuilder
from neptune.exceptions import (
    InactiveRunException,
    MetadataInconsistency,
    NeptunePossibleLegacyUsageException,
)
from neptune.handler import Handler
from neptune.internal.state import ContainerState as RunState
from neptune.metadata_containers import Run
from neptune.types import (
    Boolean,
    Integer,
)
from neptune.types.atoms.datetime import Datetime
from neptune.types.atoms.float import Float
from neptune.types.atoms.string import String
from neptune.types.namespace import Namespace
from neptune.types.value import Value



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["StringSeries"]

from typing import (
    TYPE_CHECKING,
    Iterable,
    List,
    Union,
)

from neptune.attributes.series.fetchable_series import FetchableSeries
from neptune.attributes.series.series import Series
from neptune.internal.backends.api_model import StringSeriesValues
from neptune.internal.operation import (
    ClearStringLog,
    LogStrings,
    Operation,
)
from neptune.internal.utils import is_collection
from neptune.internal.utils.logger import logger
from neptune.internal.utils.paths import path_to_str
from neptune.types.series.string_series import MAX_STRING_SERIES_VALUE_LENGTH
from neptune.types.series.string_series import StringSeries as StringSeriesVal

if TYPE_CHECKING:
    from neptune.metadata_containers import MetadataContainer

Val = StringSeriesVal
Data = str
LogOperation = LogStrings


class StringSeries(
    Series[Val, Data, LogOperation], FetchableSeries[StringSeriesValues], max_batch_size=10, operation_cls=LogOperation
):
    def __init__(self, container: "MetadataContainer", path: List[str]):
        super().__init__(container, path)
        self._value_truncation_occurred = False

    def _get_log_operations_from_value(
        self,
        value: Val,
    ) -> List[LogOperation]:
        if not self._value_truncation_occurred and value.truncated:
            # the first truncation
            self._value_truncation_occurred = True
            logger.warning(
                "Warning: string series '%s' value was"
                " longer than %s characters and was truncated."
                " This warning is printed only once per series.",
                path_to_str(self._path),
                MAX_STRING_SERIES_VALUE_LENGTH,
            )

        return super()._get_log_operations_from_value(value)

    def _get_clear_operation(self) -> Operation:
        return ClearStringLog(self._path)

    def _data_to_value(self, values: Iterable, **kwargs) -> Val:
        steps = kwargs.pop("steps", None)
        timestamps = kwargs.pop("timestamps", None)

        if kwargs:
            logger.warning("Warning: unexpected arguments (%s) in StringSeries", kwargs)

        return StringSeriesVal(values, steps=steps, timestamps=timestamps)

    def _is_value_type(self, value) -> bool:
        return isinstance(value, StringSeriesVal)

    def _handle_stringified_value(self, value) -> Union[List[str], str]:
        if is_collection(value.value):
            return list(map(str, value.value))
        return str(value.value)

    def fetch_last(self) -> str:
        val = self._backend.get_string_series_attribute(self._container_id, self._container_type, self._path)
        return val.last

    def _fetch_values_from_backend(self, offset, limit) -> StringSeriesValues:
        return self._backend.get_string_series_values(
            self._container_id, self._container_type, self._path, offset, limit
        )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["logger"]

import logging
import sys

LOGGER_NAME = "neptune-client"


class GrabbableStdoutHandler(logging.StreamHandler):
    """
    This class is like a StreamHandler using sys.stdout, but always uses
    whatever sys.stdout is currently set to rather than the value of
    sys.stderr at handler construction time.
    This enables neptune-client to capture stdout regardless
    of logging configuration time.
    Based on logging._StderrHandler from standard library.
    """

    def __init__(self, level=logging.NOTSET):
        logging.Handler.__init__(self, level)

    @property
    def stream(self):
        return sys.stdout


logger = logging.getLogger(LOGGER_NAME)

logger.propagate = False
logger.setLevel(level=logging.DEBUG)
stdout_handler = GrabbableStdoutHandler()
stdout_handler.setFormatter(logging.Formatter("%(message)s"))
logger.addHandler(stdout_handler)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ("SyncOperationProcessor",)

import os
from datetime import datetime
from pathlib import Path
from typing import Optional

from neptune.constants import SYNC_DIRECTORY
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import UniqueId
from neptune.internal.operation import Operation
from neptune.internal.operation_processors.operation_processor import OperationProcessor
from neptune.internal.operation_processors.operation_storage import (
    OperationStorage,
    get_container_dir,
)


class SyncOperationProcessor(OperationProcessor):
    def __init__(self, container_id: UniqueId, container_type: ContainerType, backend: NeptuneBackend):
        self._container_id = container_id
        self._container_type = container_type
        self._backend = backend
        self._operation_storage = OperationStorage(self._init_data_path(container_id, container_type))

    @staticmethod
    def _init_data_path(container_id: UniqueId, container_type: ContainerType) -> Path:
        now = datetime.now()
        process_path = f"exec-{now.timestamp()}-{now.strftime('%Y-%m-%d_%H.%M.%S.%f')}-{os.getpid()}"
        return get_container_dir(SYNC_DIRECTORY, container_id, container_type, process_path)

    def enqueue_operation(self, op: Operation, *, wait: bool) -> None:
        _, errors = self._backend.execute_operations(
            container_id=self._container_id,
            container_type=self._container_type,
            operations=[op],
            operation_storage=self._operation_storage,
        )
        if errors:
            raise errors[0]

    def wait(self):
        pass

    def flush(self):
        pass

    def start(self):
        pass

    def stop(self, seconds: Optional[float] = None) -> None:
        # Remove local files
        self._operation_storage.cleanup()

    def close(self) -> None:
        pass



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["status", "sync", "clear"]

from pathlib import Path
from typing import (
    List,
    Optional,
)

import click

from neptune.cli.clear import ClearRunner
from neptune.cli.path_option import path_option
from neptune.cli.status import StatusRunner
from neptune.cli.sync import SyncRunner
from neptune.common.exceptions import NeptuneException  # noqa: F401
from neptune.exceptions import (  # noqa: F401
    CannotSynchronizeOfflineRunsWithoutProject,
    ProjectNotFound,
    RunNotFound,
)
from neptune.internal.backends.api_model import (  # noqa: F401
    ApiExperiment,
    Project,
)
from neptune.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.internal.backends.neptune_backend import NeptuneBackend  # noqa: F401
from neptune.internal.credentials import Credentials
from neptune.internal.disk_queue import DiskQueue  # noqa: F401
from neptune.internal.operation import Operation  # noqa: F401


@click.command()
@path_option
def status(path: Path) -> None:
    """List synchronized and unsynchronized objects in the given directory. Trashed objects are not listed.

    Neptune stores object data on disk in the '.neptune' directory. If an object executes offline
    or if the network is unavailable as the object executes, the object data can be synchronized
    with the server with this command line utility.

    Examples:

    \b
    # List synchronized and unsynchronized objects in the current directory
    neptune status

    \b
    # List synchronized and unsynchronized objects in directory "foo/bar" without actually syncing
    neptune status --path foo/bar
    """

    status_runner = StatusRunner(backend=HostedNeptuneBackend(Credentials.from_token()))

    status_runner.synchronization_status(path)


@click.command()
@path_option
@click.option(
    "--object",
    "object_names",
    multiple=True,
    metavar="<object-name>",
    help="object name (workspace/project/short-id or UUID for offline runs) to synchronize.",
)
@click.option(
    "-p",
    "--project",
    "project_name",
    multiple=False,
    metavar="project-name",
    help="project name (workspace/project) where offline runs will be sent",
)
@click.option(
    "--offline-only",
    "offline_only",
    is_flag=True,
    default=False,
    help="synchronize only the offline runs inside '.neptune' directory",
)
def sync(
    path: Path,
    object_names: List[str],
    project_name: Optional[str],
    offline_only: Optional[bool],
):
    """Synchronizes objects with unsent data to the server.

     Neptune stores object data on disk in the '.neptune' directory. If an object executes offline
    or if the network is unavailable as the run executes,  the object data can be synchronized
    with the server with this command line utility.

    You can list unsynchronized runs with `neptune status`

    Examples:

    \b
    # Synchronize all objects in the current directory
    neptune sync

    \b
    # Synchronize all objects in the given path
    neptune sync --path foo/bar

    \b
    # Synchronize only runs "NPT-42" and "NPT-43" in "workspace/project" in the current directory
    neptune sync --object workspace/project/NPT-42 --object workspace/project/NPT-43

    \b
    # Synchronise all objects in the current directory, sending offline runs to project "workspace/project"
    neptune sync --project workspace/project

    \b
    # Synchronize only the offline run with UUID offline/a1561719-b425-4000-a65a-b5efb044d6bb
    # to project "workspace/project"
    neptune sync --project workspace/project --object offline/a1561719-b425-4000-a65a-b5efb044d6bb

    \b
    # Synchronize only the offline runs
    neptune sync --offline-only

    \b
    # Synchronize only the offline runs to project "workspace/project"
    neptune sync --project workspace/project --offline-only
    """

    backend = HostedNeptuneBackend(Credentials.from_token())
    sync_runner = SyncRunner(backend=backend)

    if offline_only:
        if object_names:
            raise click.BadParameter("--object and --offline-only are mutually exclusive")

        sync_runner.sync_all_offline_containers(path, project_name)

    elif object_names:
        sync_runner.sync_selected_containers(path, project_name, object_names)
    else:
        sync_runner.sync_all_containers(path, project_name)

    clear_runner = ClearRunner(backend=backend)

    clear_runner.clear(path, clear_eventual=False)


@click.command()
@path_option
def clear(path: Path):
    """
    Clears metadata that has been synchronized or trashed, but is still present in local storage.

    Lists objects and data to be cleared before deleting the data.

    Examples:

    \b
    # Clear junk metadata from local storage
    neptune clear

    \b
    # Clear junk metadata from directory "foo/bar"
    neptune clear --path foo/bar
    """
    backend = HostedNeptuneBackend(Credentials.from_token())
    clear_runner = ClearRunner(backend=backend)

    clear_runner.clear(path)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
import os
from dataclasses import dataclass
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
)

from neptune.common.exceptions import (
    InternalClientError,
    NeptuneException,
)
from neptune.exceptions import MalformedOperation
from neptune.internal.container_type import ContainerType
from neptune.internal.operation_processors.operation_storage import OperationStorage
from neptune.internal.types.file_types import FileType
from neptune.types.atoms.file import File

if TYPE_CHECKING:
    from neptune.attributes.attribute import Attribute
    from neptune.internal.backends.neptune_backend import NeptuneBackend
    from neptune.internal.operation_visitor import OperationVisitor

Ret = TypeVar("Ret")
T = TypeVar("T")


def all_subclasses(cls):
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in all_subclasses(c)])


@dataclass
class Operation(abc.ABC):

    path: List[str]

    @abc.abstractmethod
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        pass

    def clean(self, operation_storage: OperationStorage):
        pass

    def to_dict(self) -> dict:
        return {"type": self.__class__.__name__, "path": self.path}

    @staticmethod
    def from_dict(data: dict) -> "Operation":
        if "type" not in data:
            raise ValueError("Malformed operation {} - type is missing".format(data))
        sub_classes = {cls.__name__: cls for cls in all_subclasses(Operation)}
        if not data["type"] in sub_classes:
            raise ValueError("Malformed operation {} - unknown type {}".format(data, data["type"]))
        return sub_classes[data["type"]].from_dict(data)


@dataclass
class AssignFloat(Operation):

    value: float

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_float(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignFloat":
        return AssignFloat(data["path"], data["value"])


@dataclass
class AssignInt(Operation):

    value: int

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_int(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignInt":
        return AssignInt(data["path"], data["value"])


@dataclass
class AssignBool(Operation):

    value: bool

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_bool(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignBool":
        return AssignBool(data["path"], data["value"])


@dataclass
class AssignString(Operation):

    value: str

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_string(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignString":
        return AssignString(data["path"], data["value"])


@dataclass
class AssignDatetime(Operation):

    value: datetime

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_datetime(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = int(1000 * self.value.timestamp())
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignDatetime":
        return AssignDatetime(data["path"], datetime.fromtimestamp(data["value"] / 1000))


@dataclass
class AssignArtifact(Operation):

    hash: str

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_artifact(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["hash"] = self.hash
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignArtifact":
        return AssignArtifact(data["path"], str(data["hash"]))


@dataclass
class UploadFile(Operation):

    ext: str
    file_path: str = None
    tmp_file_name: str = None
    # `clean_after_upload` is for backward compatibility and should be removed in the future
    clean_after_upload: bool = False

    @classmethod
    def of_file(cls, value: File, attribute_path: List[str], operation_storage: OperationStorage):
        if value.file_type is FileType.LOCAL_FILE:
            operation = UploadFile(
                path=attribute_path,
                ext=value.extension,
                file_path=os.path.abspath(value.path),
            )
        elif value.file_type in (FileType.IN_MEMORY, FileType.STREAM):
            tmp_file_name = cls.get_tmp_file_name(attribute_path, value.extension)
            value._save(operation_storage.upload_path / tmp_file_name)
            operation = UploadFile(path=attribute_path, ext=value.extension, tmp_file_name=tmp_file_name)
        else:
            raise ValueError(f"Unexpected FileType: {value.file_type}")
        return operation

    def clean(self, operation_storage: OperationStorage):
        if self.clean_after_upload or self.tmp_file_name:
            os.remove(self.get_absolute_path(operation_storage))

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_upload_file(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["ext"] = self.ext
        ret["file_path"] = self.file_path
        ret["tmp_file_name"] = self.tmp_file_name
        ret["clean_after_upload"] = self.clean_after_upload
        return ret

    @staticmethod
    def from_dict(data: dict) -> "UploadFile":
        return UploadFile(
            data["path"],
            data["ext"],
            data.get("file_path"),
            data.get("tmp_file_name"),
            data.get("clean_after_upload", False),
        )

    @staticmethod
    def get_tmp_file_name(attribute_path: List[str], extension: str):
        now = datetime.now()
        tmp_file_name = (
            f"{'_'.join(attribute_path)}-{now.timestamp()}-{now.strftime('%Y-%m-%d_%H.%M.%S.%f')}.{extension}"
        )
        return tmp_file_name

    def get_absolute_path(self, operation_storage: OperationStorage) -> str:
        if self.file_path:
            return self.file_path
        elif self.tmp_file_name:
            return str(operation_storage.upload_path / self.tmp_file_name)

        raise NeptuneException("Expected 'file_path' or 'tmp_file_name' to be filled.")


@dataclass
class UploadFileContent(Operation):

    ext: str
    file_content: str

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_upload_file_content(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["ext"] = self.ext
        ret["file_content"] = self.file_content
        return ret

    @staticmethod
    def from_dict(data: dict) -> "UploadFileContent":
        return UploadFileContent(data["path"], data["ext"], data["file_content"])


@dataclass
class UploadFileSet(Operation):

    file_globs: List[str]
    reset: bool

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_upload_file_set(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["file_globs"] = self.file_globs
        ret["reset"] = str(self.reset)
        return ret

    @staticmethod
    def from_dict(data: dict) -> "UploadFileSet":
        return UploadFileSet(data["path"], data["file_globs"], data["reset"] != str(False))


class LogOperation(Operation, abc.ABC):
    pass


@dataclass
class LogSeriesValue(Generic[T]):

    value: T
    step: Optional[float]
    ts: float

    def to_dict(self, value_serializer=lambda x: x) -> dict:
        return {"value": value_serializer(self.value), "step": self.step, "ts": self.ts}

    @staticmethod
    def from_dict(data: dict, value_deserializer=lambda x: x) -> "LogSeriesValue[T]":
        return LogSeriesValue[T](value_deserializer(data["value"]), data.get("step", None), data["ts"])


@dataclass
class LogFloats(LogOperation):

    ValueType = LogSeriesValue[float]

    values: List[ValueType]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_log_floats(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = [value.to_dict() for value in self.values]
        return ret

    @staticmethod
    def from_dict(data: dict) -> "LogFloats":
        return LogFloats(
            data["path"],
            [LogFloats.ValueType.from_dict(value) for value in data["values"]],
        )


@dataclass
class LogStrings(LogOperation):

    ValueType = LogSeriesValue[str]

    values: List[ValueType]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_log_strings(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = [value.to_dict() for value in self.values]
        return ret

    @staticmethod
    def from_dict(data: dict) -> "LogStrings":
        return LogStrings(
            data["path"],
            [LogStrings.ValueType.from_dict(value) for value in data["values"]],
        )


@dataclass
class ImageValue:
    data: Optional[str]
    name: Optional[str]
    description: Optional[str]

    @staticmethod
    def serializer(obj: "ImageValue"):
        return dict(data=obj.data, name=obj.name, description=obj.description)

    @staticmethod
    def deserializer(obj) -> "ImageValue":
        if obj is None:
            return ImageValue(None, None, None)
        if isinstance(obj, str):
            return ImageValue(data=obj, name=None, description=None)
        if isinstance(obj, dict):
            return ImageValue(data=obj["data"], name=obj["name"], description=obj["description"])
        else:
            raise InternalClientError("Run data on disk is malformed or was saved by newer version of Neptune Library")


@dataclass
class LogImages(LogOperation):

    ValueType = LogSeriesValue[ImageValue]

    values: List[ValueType]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_log_images(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = [value.to_dict(ImageValue.serializer) for value in self.values]
        return ret

    @staticmethod
    def from_dict(data: dict) -> "LogImages":
        return LogImages(
            data["path"],
            [LogImages.ValueType.from_dict(value, ImageValue.deserializer) for value in data["values"]],
        )


@dataclass
class ClearFloatLog(Operation):
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_clear_float_log(self)

    @staticmethod
    def from_dict(data: dict) -> "ClearFloatLog":
        return ClearFloatLog(data["path"])


@dataclass
class ClearStringLog(Operation):
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_clear_string_log(self)

    @staticmethod
    def from_dict(data: dict) -> "ClearStringLog":
        return ClearStringLog(data["path"])


@dataclass
class ClearImageLog(Operation):
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_clear_image_log(self)

    @staticmethod
    def from_dict(data: dict) -> "ClearImageLog":
        return ClearImageLog(data["path"])


@dataclass
class ConfigFloatSeries(Operation):

    min: Optional[float]
    max: Optional[float]
    unit: Optional[str]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_config_float_series(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["min"] = self.min
        ret["max"] = self.max
        ret["unit"] = self.unit
        return ret

    @staticmethod
    def from_dict(data: dict) -> "ConfigFloatSeries":
        return ConfigFloatSeries(data["path"], data["min"], data["max"], data["unit"])


@dataclass
class AddStrings(Operation):

    values: Set[str]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_add_strings(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = list(self.values)
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AddStrings":
        return AddStrings(data["path"], set(data["values"]))


@dataclass
class RemoveStrings(Operation):

    values: Set[str]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_remove_strings(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = list(self.values)
        return ret

    @staticmethod
    def from_dict(data: dict) -> "RemoveStrings":
        return RemoveStrings(data["path"], set(data["values"]))


@dataclass
class ClearStringSet(Operation):
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_clear_string_set(self)

    @staticmethod
    def from_dict(data: dict) -> "ClearStringSet":
        return ClearStringSet(data["path"])


@dataclass
class DeleteFiles(Operation):

    file_paths: Set[str]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_delete_files(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["file_paths"] = list(self.file_paths)
        return ret

    @staticmethod
    def from_dict(data: dict) -> "DeleteFiles":
        return DeleteFiles(data["path"], set(data["file_paths"]))


@dataclass
class DeleteAttribute(Operation):
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_delete_attribute(self)

    @staticmethod
    def from_dict(data: dict) -> "DeleteAttribute":
        return DeleteAttribute(data["path"])


@dataclass
class TrackFilesToArtifact(Operation):
    project_id: str
    entries: List[Tuple[str, Optional[str]]]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_track_files_to_artifact(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["entries"] = self.entries
        ret["project_id"] = self.project_id
        return ret

    @staticmethod
    def from_dict(data: dict) -> "TrackFilesToArtifact":
        return TrackFilesToArtifact(
            path=data["path"],
            project_id=data["project_id"],
            entries=list(map(tuple, data["entries"])),
        )


@dataclass
class ClearArtifact(Operation):
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_clear_artifact(self)

    @staticmethod
    def from_dict(data: dict) -> "ClearArtifact":
        return ClearArtifact(data["path"])


@dataclass
class CopyAttribute(Operation):
    container_id: str
    container_type: ContainerType
    source_path: List[str]
    source_attr_cls: Type["Attribute"]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_copy_attribute(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["container_id"] = self.container_id
        ret["container_type"] = self.container_type.value
        ret["source_path"] = self.source_path
        ret["source_attr_name"] = self.source_attr_cls.__name__
        return ret

    @staticmethod
    def from_dict(data: dict) -> "CopyAttribute":
        from neptune.attributes.attribute import Attribute

        source_attr_cls = {cls.__name__: cls for cls in all_subclasses(Attribute) if cls.supports_copy}.get(
            data["source_attr_name"]
        )

        if source_attr_cls is None:
            raise MalformedOperation("Copy of non-copiable type found in queue!")

        return CopyAttribute(
            data["path"],
            data["container_id"],
            ContainerType(data["container_type"]),
            data["source_path"],
            source_attr_cls,
        )

    def resolve(self, backend: "NeptuneBackend") -> Operation:
        # repack CopyAttribute op into target attribute assignment
        getter = self.source_attr_cls.getter
        create_assignment_operation = self.source_attr_cls.create_assignment_operation
        value = getter(backend, self.container_id, self.container_type, self.source_path)
        return create_assignment_operation(self.path, value)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["OperationApiObjectConverter"]

from neptune.common.exceptions import InternalClientError
from neptune.internal.operation import (
    AddStrings,
    AssignArtifact,
    AssignBool,
    AssignDatetime,
    AssignFloat,
    AssignInt,
    AssignString,
    ClearArtifact,
    ClearFloatLog,
    ClearImageLog,
    ClearStringLog,
    ClearStringSet,
    ConfigFloatSeries,
    CopyAttribute,
    DeleteAttribute,
    DeleteFiles,
    LogFloats,
    LogImages,
    LogStrings,
    Operation,
    RemoveStrings,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent,
    UploadFileSet,
)
from neptune.internal.operation_visitor import (
    OperationVisitor,
    Ret,
)


class OperationApiObjectConverter(OperationVisitor[dict]):
    def convert(self, op: Operation) -> dict:
        return op.accept(self)

    def visit_assign_float(self, op: AssignFloat) -> dict:
        return {"value": op.value}

    def visit_assign_int(self, op: AssignInt) -> dict:
        return {"value": op.value}

    def visit_assign_bool(self, op: AssignBool) -> dict:
        return {"value": op.value}

    def visit_assign_string(self, op: AssignString) -> dict:
        return {"value": op.value}

    def visit_assign_datetime(self, op: AssignDatetime) -> Ret:
        return {"valueMilliseconds": int(1000 * op.value.timestamp())}

    def visit_assign_artifact(self, op: AssignArtifact) -> dict:
        return {"hash": op.hash}

    def visit_upload_file(self, _: UploadFile) -> dict:
        raise InternalClientError("Specialized endpoint should be used to upload file attribute")

    def visit_upload_file_content(self, _: UploadFileContent) -> dict:
        raise InternalClientError("Specialized endpoint should be used to upload file attribute")

    def visit_upload_file_set(self, op: UploadFileSet) -> Ret:
        raise InternalClientError("Specialized endpoints should be used to upload file set attribute")

    def visit_log_floats(self, op: LogFloats) -> dict:
        return {
            "entries": [
                {
                    "value": value.value,
                    "step": value.step,
                    "timestampMilliseconds": int(value.ts * 1000),
                }
                for value in op.values
            ]
        }

    def visit_log_strings(self, op: LogStrings) -> dict:
        return {
            "entries": [
                {
                    "value": value.value,
                    "step": value.step,
                    "timestampMilliseconds": int(value.ts * 1000),
                }
                for value in op.values
            ]
        }

    def visit_log_images(self, op: LogImages) -> dict:
        return {
            "entries": [
                {
                    "value": {
                        "data": value.value.data,
                        "name": value.value.name,
                        "description": value.value.description,
                    },
                    "step": value.step,
                    "timestampMilliseconds": int(value.ts * 1000),
                }
                for value in op.values
            ]
        }

    def visit_clear_float_log(self, _: ClearFloatLog) -> dict:
        return {}

    def visit_clear_string_log(self, _: ClearStringLog) -> dict:
        return {}

    def visit_clear_image_log(self, _: ClearImageLog) -> dict:
        return {}

    def visit_config_float_series(self, op: ConfigFloatSeries) -> dict:
        return {"min": op.min, "max": op.max, "unit": op.unit}

    def visit_add_strings(self, op: AddStrings) -> dict:
        return {"values": list(op.values)}

    def visit_remove_strings(self, op: RemoveStrings) -> dict:
        return {"values": list(op.values)}

    def visit_delete_attribute(self, _: DeleteAttribute) -> dict:
        return {}

    def visit_clear_string_set(self, _: ClearStringSet) -> dict:
        return {}

    def visit_delete_files(self, op: DeleteFiles) -> Ret:
        return {"filePaths": list(op.file_paths)}

    def visit_track_files_to_artifact(self, op: TrackFilesToArtifact) -> dict:
        raise InternalClientError("Specialized endpoint should be used to track artifact files")

    def visit_clear_artifact(self, _: ClearArtifact) -> Ret:
        return {}

    def visit_copy_attribute(self, _: CopyAttribute) -> Ret:
        raise NotImplementedError("This operation is client-side only")



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["ContainersManager"]

import abc
from pathlib import Path
from typing import (
    List,
    Tuple,
)

from neptune.cli.utils import (
    get_metadata_container,
    is_container_synced_and_remove_junk,
    iterate_containers,
)
from neptune.constants import (
    ASYNC_DIRECTORY,
    OFFLINE_DIRECTORY,
    SYNC_DIRECTORY,
)
from neptune.internal.backends.api_model import ApiExperiment
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.id_formats import UniqueId


class ContainersManager(abc.ABC):
    _backend: NeptuneBackend

    def __init__(self, backend: NeptuneBackend, base_path: Path):
        self._backend = backend
        self._base_path = base_path

    def partition_containers_and_clean_junk(
        self,
        base_path: Path,
    ) -> Tuple[List[ApiExperiment], List[ApiExperiment], List[Path]]:
        synced_containers = []
        unsynced_containers = []
        not_found = []
        async_path = base_path / ASYNC_DIRECTORY
        for container_type, container_id, path in iterate_containers(async_path):
            metadata_container = get_metadata_container(
                backend=self._backend,
                container_id=container_id,
                container_type=container_type,
            )
            if metadata_container:
                if is_container_synced_and_remove_junk(path):
                    synced_containers.append(metadata_container)
                else:

                    unsynced_containers.append(metadata_container)
            else:
                not_found.append(path)

        synced_containers = [obj for obj in synced_containers if obj]
        unsynced_containers = [obj for obj in unsynced_containers if obj]

        return synced_containers, unsynced_containers, not_found

    def resolve_async_path(self, container: ApiExperiment) -> Path:
        return self._base_path / ASYNC_DIRECTORY / container.type.create_dir_name(container.id)

    def resolve_offline_container_dir(self, offline_id: UniqueId):
        return self._base_path / OFFLINE_DIRECTORY / offline_id

    def iterate_sync_containers(self):
        return iterate_containers(self._base_path / SYNC_DIRECTORY)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["BackgroundJobList"]

import time
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
)

from neptune.internal.background_job import BackgroundJob

if TYPE_CHECKING:
    from neptune.metadata_containers import MetadataContainer


class BackgroundJobList(BackgroundJob):
    def __init__(self, jobs: List[BackgroundJob]):
        self._jobs = jobs

    def start(self, container: "MetadataContainer"):
        for job in self._jobs:
            job.start(container)

    def stop(self):
        for job in self._jobs:
            job.stop()

    def join(self, seconds: Optional[float] = None):
        ts = time.time()
        for job in self._jobs:
            sec_left = None if seconds is None else seconds - (time.time() - ts)
            job.join(sec_left)

    def pause(self):
        for job in self._jobs:
            job.pause()

    def resume(self):
        for job in self._jobs:
            job.resume()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "FileComposite",
    "LocalFileComposite",
    "InMemoryComposite",
    "FileComposite",
    "StreamComposite",
]

import abc
import enum
import io
import os
from functools import wraps
from io import IOBase
from typing import (
    Optional,
    Union,
)

from neptune.common.exceptions import NeptuneException
from neptune.exceptions import StreamAlreadyUsedException
from neptune.internal.utils import verify_type


class FileType(enum.Enum):
    LOCAL_FILE = "LOCAL_FILE"
    IN_MEMORY = "IN_MEMORY"
    STREAM = "STREAM"


class FileComposite(abc.ABC):
    """
    Composite class defining behaviour of neptune.types.atoms.file.File
    """

    file_type: FileType = None

    def __init__(self, extension: str):
        verify_type("extension", extension, str)
        self._extension = extension

    @property
    def extension(self):
        return self._extension

    @property
    def path(self):
        raise NeptuneException(f"`path` attribute is not supported for {self.file_type}")

    @property
    def content(self):
        raise NeptuneException(f"`content` attribute is not supported for {self.file_type}")

    def save(self, path):
        raise NeptuneException(f"`save` method is not supported for {self.file_type}")


class LocalFileComposite(FileComposite):
    file_type = FileType.LOCAL_FILE

    def __init__(self, path: str, extension: Optional[str] = None):
        try:
            ext = os.path.splitext(path)[1]
            ext = ext[1:] if ext else ""
        except ValueError:
            ext = ""
        super().__init__(extension or ext)

        self._path = path

    @property
    def path(self):
        return self._path

    def __str__(self):
        return f"File(path={self.path})"


class InMemoryComposite(FileComposite):
    file_type = FileType.IN_MEMORY

    def __init__(self, content: Union[str, bytes], extension: Optional[str] = None):
        if isinstance(content, str):
            ext = "txt"
            content = content.encode("utf-8")
        else:
            ext = "bin"
        super().__init__(extension or ext)

        self._content = content

    @property
    def content(self):
        return self._content

    def save(self, path):
        with open(path, "wb") as f:
            f.write(self._content)

    def __str__(self):
        return "File(content=...)"


def read_once(f):
    """Decorator for validating read once on STREAM objects"""

    @wraps(f)
    def func(self: "StreamComposite", *args, **kwargs):
        if self._stream_read:
            raise StreamAlreadyUsedException()
        self._stream_read = True
        return f(self, *args, **kwargs)

    return func


class StreamComposite(FileComposite):
    file_type = FileType.STREAM

    def __init__(self, stream: IOBase, seek: Optional[int] = 0, extension: Optional[str] = None):
        verify_type("stream", stream, (IOBase, type(None)))
        verify_type("extension", extension, (str, type(None)))

        if seek is not None and stream.seekable():
            stream.seek(seek)
        if extension is None:
            extension = "txt" if isinstance(stream, io.TextIOBase) else "bin"
        super().__init__(extension)

        self._stream = stream
        self._stream_read = False

    @property
    @read_once
    def content(self):
        val = self._stream.read()
        if isinstance(self._stream, io.TextIOBase):
            val = val.encode()
        return val

    @read_once
    def save(self, path):
        with open(path, "wb") as f:
            buffer_ = self._stream.read(io.DEFAULT_BUFFER_SIZE)
            while buffer_:
                # TODO: replace with Walrus Operator once python3.7 support is dropped
                if isinstance(self._stream, io.TextIOBase):
                    buffer_ = buffer_.encode()
                f.write(buffer_)
                buffer_ = self._stream.read(io.DEFAULT_BUFFER_SIZE)

    def __str__(self):
        return f"File(stream={self._stream})"



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["CopiableAtom"]

import abc
import typing

from neptune.attributes.atoms.atom import Atom
from neptune.internal.container_type import ContainerType
from neptune.internal.operation import CopyAttribute
from neptune.internal.utils.paths import parse_path
from neptune.types.value_copy import ValueCopy

if typing.TYPE_CHECKING:
    from neptune.internal.backends.neptune_backend import NeptuneBackend


class CopiableAtom(Atom):
    supports_copy = True

    def copy(self, value: ValueCopy, *, wait: bool = False):
        with self._container.lock():
            source_path = value.source_handler._path
            source_attr = value.source_handler._get_attribute()
            self._enqueue_operation(
                CopyAttribute(
                    self._path,
                    container_id=source_attr._container_id,
                    container_type=source_attr._container_type,
                    source_path=parse_path(source_path),
                    source_attr_cls=source_attr.__class__,
                ),
                wait=wait,
            )

    @staticmethod
    @abc.abstractmethod
    def create_assignment_operation(path, value: int):
        ...

    @staticmethod
    @abc.abstractmethod
    def getter(
        backend: "NeptuneBackend",
        container_id: str,
        container_type: ContainerType,
        path: typing.List[str],
    ) -> int:
        ...

    def fetch(self):
        return self.getter(self._backend, self._container_id, self._container_type, self._path)



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["SupportsNamespaces", "NeptuneObject", "NeptuneObjectCallback"]

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    Union,
)

if TYPE_CHECKING:
    from neptune.handler import Handler


class SupportsNamespaces(ABC):
    """
    Interface for Neptune objects that supports subscripting (selecting namespaces)
    It could be a Run, Model, ModelVersion, Project or already selected namespace (Handler).

    Example:
        >>> from neptune import init_run
        >>> from neptune.typing import SupportsNamespaces
        >>> class NeptuneCallback:
        ...     # Proper type hinting of `start_from` parameter.
        ...     def __init__(self, start_from: SupportsNamespaces):
        ...         self._start_from = start_from
        ...
        ...     def log_accuracy(self, accuracy: float) -> None:
        ...         self._start_from["train/acc"] = accuracy
        ...
        >>> run = init_run()
        >>> callback = NeptuneCallback(start_from=run)
        >>> callback.log_accuracy(0.8)
        >>> # or
        ... callback = NeptuneCallback(start_from=run["some/random/path"])
        >>> callback.log_accuracy(0.8)
    """

    @abstractmethod
    def __getitem__(self, path: str) -> "Handler":
        ...

    @abstractmethod
    def __setitem__(self, key: str, value) -> None:
        ...

    @abstractmethod
    def __delitem__(self, path) -> None:
        ...

    @abstractmethod
    def get_root_object(self) -> "SupportsNamespaces":
        ...


class NeptuneObject(SupportsNamespaces, ABC):
    @abstractmethod
    def stop(self, *, seconds: Optional[Union[float, int]] = None) -> None:
        ...


NeptuneObjectCallback = Callable[[NeptuneObject], None]



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "upload_file_attribute",
    "upload_file_set_attribute",
    "get_unique_upload_entries",
    "download_file_set_attribute",
]

import collections
import enum
import json
import os
import time
from io import BytesIO
from typing import (
    AnyStr,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Union,
)
from urllib.parse import urlencode

from bravado.exception import (
    HTTPPaymentRequired,
    HTTPUnprocessableEntity,
)
from bravado.requests_client import RequestsClient
from requests import (
    Request,
    Response,
)

from neptune.common.backends.api_model import MultipartConfig
from neptune.common.backends.utils import with_api_exceptions_handler
from neptune.common.exceptions import (
    InternalClientError,
    NeptuneException,
    UploadedFileChanged,
)
from neptune.common.hardware.constants import BYTES_IN_ONE_MB
from neptune.common.storage.datastream import (
    FileChunk,
    FileChunker,
    compress_to_tar_gz_in_memory,
)
from neptune.common.storage.storage_utils import (
    AttributeUploadConfiguration,
    UploadEntry,
    normalize_file_name,
    scan_unique_upload_entries,
    split_upload_files,
)
from neptune.exceptions import (
    FileUploadError,
    MetadataInconsistency,
    NeptuneLimitExceedException,
)
from neptune.internal.backends.swagger_client_wrapper import (
    ApiMethodWrapper,
    SwaggerClientWrapper,
)
from neptune.internal.backends.utils import (
    build_operation_url,
    handle_server_raw_response_messages,
)
from neptune.internal.utils import (
    get_absolute_paths,
    get_common_root,
)
from neptune.internal.utils.logger import logger

DEFAULT_CHUNK_SIZE = 5 * BYTES_IN_ONE_MB
DEFAULT_UPLOAD_CONFIG = AttributeUploadConfiguration(chunk_size=DEFAULT_CHUNK_SIZE)


class FileUploadTarget(enum.Enum):
    FILE_ATOM = "file"
    FILE_SET = "fileset"


def upload_file_attribute(
    swagger_client: SwaggerClientWrapper,
    container_id: str,
    attribute: str,
    source: Union[str, bytes],
    ext: str,
    multipart_config: Optional[MultipartConfig],
) -> List[NeptuneException]:
    if isinstance(source, str) and not os.path.isfile(source):
        return [FileUploadError(source, "Path not found or is a not a file.")]

    target = attribute
    if ext:
        target += "." + ext

    try:
        upload_entry = UploadEntry(source if isinstance(source, str) else BytesIO(source), target)
        _multichunk_upload_with_retry(
            upload_entry,
            query_params={
                "experimentIdentifier": container_id,
                "attribute": attribute,
                "ext": ext,
            },
            swagger_client=swagger_client,
            multipart_config=multipart_config,
            target=FileUploadTarget.FILE_ATOM,
        )
    except MetadataInconsistency as e:
        return [e]


def upload_file_set_attribute(
    swagger_client: SwaggerClientWrapper,
    container_id: str,
    attribute: str,
    file_globs: Iterable[str],
    reset: bool,
    multipart_config: Optional[MultipartConfig],
) -> List[NeptuneException]:
    unique_upload_entries = get_unique_upload_entries(file_globs)

    try:
        upload_configuration = DEFAULT_UPLOAD_CONFIG
        for package in split_upload_files(
            upload_entries=unique_upload_entries,
            upload_configuration=upload_configuration,
        ):
            if package.is_empty() and not reset:
                continue

            uploading_multiple_entries = package.len > 1
            creating_a_single_empty_dir = (
                package.len == 1 and not package.items[0].is_stream() and os.path.isdir(package.items[0].source)
            )

            if uploading_multiple_entries or creating_a_single_empty_dir or package.is_empty():
                data = compress_to_tar_gz_in_memory(upload_entries=package.items)
                url = build_operation_url(
                    swagger_client.swagger_spec.api_url,
                    swagger_client.api.uploadFileSetAttributeTar.operation.path_name,
                )
                result = upload_raw_data(
                    http_client=swagger_client.swagger_spec.http_client,
                    url=url,
                    data=data,
                    headers={"Content-Type": "application/octet-stream"},
                    query_params={
                        "experimentId": container_id,
                        "attribute": attribute,
                        "reset": str(reset),
                    },
                )
                _attribute_upload_response_handler(result)
            else:
                upload_entry = package.items[0]
                _multichunk_upload_with_retry(
                    upload_entry,
                    query_params={
                        "experimentIdentifier": container_id,
                        "attribute": attribute,
                        "subPath": upload_entry.target_path,
                    },
                    swagger_client=swagger_client,
                    multipart_config=multipart_config,
                    target=FileUploadTarget.FILE_SET,
                )

            reset = False
    except MetadataInconsistency as e:
        if len(e.args) == 1:
            return [e]
        else:
            return [MetadataInconsistency(desc) for desc in e.args]


def get_unique_upload_entries(file_globs: Iterable[str]) -> Set[UploadEntry]:
    absolute_paths = get_absolute_paths(file_globs)
    common_root = get_common_root(absolute_paths)

    upload_entries: List[UploadEntry] = []
    if common_root is not None:
        for absolute_path in absolute_paths:
            upload_entries.append(
                UploadEntry(
                    absolute_path,
                    normalize_file_name(os.path.relpath(absolute_path, common_root)),
                )
            )
    else:
        for absolute_path in absolute_paths:
            upload_entries.append(UploadEntry(absolute_path, normalize_file_name(absolute_path)))

    return scan_unique_upload_entries(upload_entries)


def _attribute_upload_response_handler(result: bytes) -> None:
    try:
        parsed = json.loads(result)
    except json.JSONDecodeError:
        raise InternalClientError("Unexpected response from server: {}".format(result))

    if isinstance(parsed, type(None)):
        # old format with empty optional error
        return
    if isinstance(parsed, dict):
        if "errorDescription" in parsed:
            # old format with optional error
            raise MetadataInconsistency(parsed["errorDescription"])
        elif "errors" in parsed:
            # new format with a list of errors
            error_list = parsed["errors"]
            if isinstance(error_list, list):
                if len(error_list) == 0:
                    return
                try:
                    raise MetadataInconsistency(*[item["errorDescription"] for item in parsed["errors"]])
                except KeyError:
                    # fall into default InternalClientError
                    pass

    raise InternalClientError("Unexpected response from server: {}".format(result))


MultipartUrlSet = collections.namedtuple("MultipartUrlSet", ["start_chunked", "finish_chunked", "send_chunk", "single"])

MULTIPART_URLS = {
    FileUploadTarget.FILE_ATOM: MultipartUrlSet(
        "fileAtomMultipartUploadStart",
        "fileAtomMultipartUploadFinish",
        "fileAtomMultipartUploadPart",
        "fileAtomUpload",
    ),
    FileUploadTarget.FILE_SET: MultipartUrlSet(
        "fileSetFileMultipartUploadStart",
        "fileSetFileMultipartUploadFinish",
        "fileSetFileMultipartUploadPart",
        "fileSetFileUpload",
    ),
}


def _build_multipart_urlset(swagger_client: SwaggerClientWrapper, target: FileUploadTarget) -> MultipartUrlSet:
    urlnameset = MULTIPART_URLS[target]
    return MultipartUrlSet(
        start_chunked=with_api_exceptions_handler(getattr(swagger_client.api, urlnameset.start_chunked)),
        finish_chunked=with_api_exceptions_handler(getattr(swagger_client.api, urlnameset.finish_chunked)),
        send_chunk=build_operation_url(
            swagger_client.swagger_spec.api_url,
            getattr(swagger_client.api, urlnameset.send_chunk).operation.path_name,
        ),
        single=build_operation_url(
            swagger_client.swagger_spec.api_url,
            getattr(swagger_client.api, urlnameset.single).operation.path_name,
        ),
    )


def _multichunk_upload_with_retry(
    upload_entry: UploadEntry,
    swagger_client: SwaggerClientWrapper,
    query_params: dict,
    multipart_config: Optional[MultipartConfig],
    target: FileUploadTarget,
):
    urlset = _build_multipart_urlset(swagger_client, target)
    while True:
        try:
            return _multichunk_upload(upload_entry, swagger_client, query_params, multipart_config, urlset)
        except UploadedFileChanged as e:
            logger.error(str(e))


def _multichunk_upload(
    upload_entry: UploadEntry,
    swagger_client: SwaggerClientWrapper,
    query_params: dict,
    multipart_config: Optional[MultipartConfig],
    urlset: MultipartUrlSet,
):
    if multipart_config is None:
        multipart_config = MultipartConfig.get_default()

    file_stream = upload_entry.get_stream()
    entry_length = upload_entry.length()
    try:
        if entry_length <= multipart_config.max_single_part_size:
            # single upload
            data = file_stream.read()
            result = upload_raw_data(
                http_client=swagger_client.swagger_spec.http_client,
                url=urlset.single,
                data=data,
                query_params=query_params,
            )
            _attribute_upload_response_handler(result)
        else:
            # chunked upload
            result = urlset.start_chunked(**query_params, totalLength=entry_length).response().result
            if result.errors:
                raise MetadataInconsistency([err.errorDescription for err in result.errors])

            no_ext_query_params = query_params.copy()
            if "ext" in no_ext_query_params:
                del no_ext_query_params["ext"]

            upload_id = result.uploadId
            chunker = FileChunker(
                None if upload_entry.is_stream() else upload_entry.source,
                file_stream,
                entry_length,
                multipart_config,
            )
            for idx, chunk in enumerate(chunker.generate()):
                result = upload_raw_data(
                    http_client=swagger_client.swagger_spec.http_client,
                    url=urlset.send_chunk,
                    data=chunk.data,
                    headers={"X-Range": _build_x_range(chunk, entry_length)},
                    query_params={
                        "uploadId": upload_id,
                        "uploadPartIdx": idx,
                        **no_ext_query_params,
                    },
                )
                _attribute_upload_response_handler(result)

            result = urlset.finish_chunked(**no_ext_query_params, uploadId=upload_id).response().result
            if result.errors:
                raise MetadataInconsistency([err.errorDescription for err in result.errors])
        return []
    finally:
        file_stream.close()


def _build_x_range(chunk: FileChunk, total_size: int) -> str:
    return "bytes=%d-%d/%d" % (
        chunk.start,
        chunk.end - 1,
        total_size,
    )


@with_api_exceptions_handler
def upload_raw_data(
    http_client: RequestsClient,
    url: str,
    data: AnyStr,
    path_params: Optional[Dict[str, str]] = None,
    query_params: Optional[Dict[str, str]] = None,
    headers: Optional[Dict[str, str]] = None,
):
    url = _generate_url(url=url, path_params=path_params, query_params=query_params)

    session = http_client.session
    request = http_client.authenticator.apply(Request(method="POST", url=url, data=data, headers=headers))
    response = handle_server_raw_response_messages(session.send(session.prepare_request(request)))

    if response.status_code >= 300:
        ApiMethodWrapper.handle_neptune_http_errors(response)
    if response.status_code in (
        HTTPUnprocessableEntity.status_code,
        HTTPPaymentRequired.status_code,
    ):
        raise NeptuneLimitExceedException(reason=response.json().get("title", "Unknown reason"))
    response.raise_for_status()

    return response.content


def download_image_series_element(
    swagger_client: SwaggerClientWrapper,
    container_id: str,
    attribute: str,
    index: int,
    destination: str,
):
    url = build_operation_url(
        swagger_client.swagger_spec.api_url,
        swagger_client.api.getImageSeriesValue.operation.path_name,
    )
    response = _download_raw_data(
        http_client=swagger_client.swagger_spec.http_client,
        url=url,
        headers={},
        query_params={
            "experimentId": container_id,
            "attribute": attribute,
            "index": index,
        },
    )
    _store_response_as_file(
        response,
        os.path.join(
            destination,
            "{}.{}".format(index, response.headers["content-type"].split("/")[-1]),
        ),
    )


def download_file_attribute(
    swagger_client: SwaggerClientWrapper,
    container_id: str,
    attribute: str,
    destination: Optional[str] = None,
):
    url = build_operation_url(
        swagger_client.swagger_spec.api_url,
        swagger_client.api.downloadAttribute.operation.path_name,
    )
    response = _download_raw_data(
        http_client=swagger_client.swagger_spec.http_client,
        url=url,
        headers={"Accept": "application/octet-stream"},
        query_params={"experimentId": container_id, "attribute": attribute},
    )
    _store_response_as_file(response, destination)


def download_file_set_attribute(
    swagger_client: SwaggerClientWrapper,
    download_id: str,
    destination: Optional[str] = None,
):
    download_url: Optional[str] = _get_download_url(swagger_client, download_id)
    next_sleep = 0.5
    while download_url is None:
        time.sleep(next_sleep)
        next_sleep = min(2 * next_sleep, 5)
        download_url = _get_download_url(swagger_client, download_id)

    response = _download_raw_data(
        http_client=swagger_client.swagger_spec.http_client,
        url=download_url,
        headers={"Accept": "application/zip"},
    )
    _store_response_as_file(response, destination)


def _get_download_url(swagger_client: SwaggerClientWrapper, download_id: str):
    params = {"id": download_id}
    download_request = swagger_client.api.getDownloadPrepareRequest(**params).response().result
    return download_request.downloadUrl


def _store_response_as_file(response: Response, destination: Optional[str] = None):
    if destination is None:
        target_file = _get_content_disposition_filename(response)
    elif os.path.isdir(destination):
        target_file = os.path.join(destination, _get_content_disposition_filename(response))
    else:
        target_file = destination
    with response:
        with open(target_file, "wb") as f:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                if chunk:
                    f.write(chunk)


def _get_content_disposition_filename(response: Response) -> str:
    content_disposition = response.headers["Content-Disposition"]
    return content_disposition[content_disposition.rfind("filename=") + 9 :].strip('"')


@with_api_exceptions_handler
def _download_raw_data(
    http_client: RequestsClient,
    url: str,
    path_params: Optional[Dict[str, str]] = None,
    query_params: Optional[Dict[str, str]] = None,
    headers: Optional[Dict[str, str]] = None,
) -> Response:
    url = _generate_url(url=url, path_params=path_params, query_params=query_params)

    session = http_client.session
    request = http_client.authenticator.apply(Request(method="GET", url=url, headers=headers))

    response = handle_server_raw_response_messages(session.send(session.prepare_request(request), stream=True))

    response.raise_for_status()
    return response


def _generate_url(
    url: str,
    path_params: Optional[Dict[str, str]] = None,
    query_params: Optional[Dict[str, str]] = None,
) -> str:
    for key, val in (path_params or dict()).items():
        url = url.replace("{" + key + "}", val)
    if query_params:
        url = url + "?" + urlencode(list(query_params.items()))
    return url
